import os.path
from collections import OrderedDict
from typing import Tuple, Union

import astropy.units as u

import exosim.log as log
import exosim.tasks.parse as parse
import exosim.utils as utils
from exosim.models.channel import Channel
from exosim.output import Output, SetOutput
from exosim.utils.ascii_arts import astronomer1
from exosim.utils.prepare_recipes import (
    clean_config_files,
    copy_input_files,
    load_options,
)
from exosim.utils.runConfig import RunConfig
from exosim.utils.timed_class import TimedClass
from exosim.utils.types import OutputType


class CreateFocalPlane(TimedClass, log.Logger):
    """
    Pipeline to create the instrument focal planes.
    This pipeline loads the configuration file and produces an output, if indicated,
    where all the products are stored.
    It loads the source SED and the foregrounds and, after the optical chain production,
    it estimates the focal plane for the source and for the foregrounds.

    Attributes
    ------------
    mainConfig: dict
        This is parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
    output: :class:`~exosim.output.output.Output` (optional)
        output file
    payloadConfig: dict
        payload configuration dictionary extracted from mainConfig`
    time: :class:`~astropy.units.Quantity`
        time grid.
    wl_grid: :class:`~numpy.ndarray` or :class:`~astropy.units.Quantity`
        wavelength grid.

    Examples
    --------

    >>> import exosim.recipes as recipes
    >>> recipes.CreateFocalPlane(options_file= 'main _configuration.xml',
    >>>                          output_file = 'output_file.h5')

    """

    def __init__(
        self,
        options_file: Union[str, dict],
        output_file: str,
        store_config: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        options_file: str or dict
            input configuration file
        output_file: str
            output file
        store_config: bool
            store the input configuration into the output (default is False)
        """
        super().__init__()

        self.graphics(astronomer1)
        RunConfig.stats()
        self.announce("started")
        clean_config_files()
        self.mainConfig, self.payloadConfig = load_options(options_file)

        if output_file is not None:
            copy_input_files(os.path.dirname(os.path.abspath(output_file)))

        self.output = SetOutput(output_file)

        # wavelength and time grid definition
        self.wl_grid = utils.grids.wl_grid(
            self.mainConfig["wl_grid"]["wl_min"],
            self.mainConfig["wl_grid"]["wl_max"],
            self.mainConfig["wl_grid"]["logbin_resolution"],
        )

        if "time_grid" in self.mainConfig.keys():
            self.time_grid = utils.grids.time_grid(
                self.mainConfig["time_grid"]["start_time"],
                self.mainConfig["time_grid"]["end_time"],
                self.mainConfig["time_grid"].get(
                    "low_frequencies_resolution", None
                ),
            )
        else:
            self.time_grid = [0] * u.hr

        if "pointing" in self.mainConfig.keys():
            pointing = (
                self.mainConfig["pointing"]["ra"],
                self.mainConfig["pointing"]["dec"],
            )
        else:
            pointing = None

        # starting the pipeline
        self.info("Focal plane pipeline started")
        with self.output.use(append=True, cache=True) as out:
            # store configuration
            if store_config:
                out.store_dictionary(self.mainConfig, "configuration")

            # common path and source preparation
            sources, common_path = self.prepare_environment(out)
            self.log_runtime("preparation ended time", "info")

            # instrument focal plane production
            ch_out = out.create_group("channels")
            if isinstance(self.payloadConfig["channel"], OrderedDict):
                for ch in self.payloadConfig["channel"].keys():
                    self.announce("channel {} started".format(ch))
                    self.run_channel(
                        self.payloadConfig["channel"][ch],
                        common_path,
                        sources,
                        pointing,
                        ch_out,
                    )
                    self.log_runtime("{} ended in".format(ch), "info")

            # else, we load the only source available
            else:
                self.run_channel(
                    self.payloadConfig["channel"],
                    common_path,
                    sources,
                    pointing,
                    ch_out,
                )

                self.log_runtime(
                    "channel {} ended time".format(
                        self.payloadConfig["channel"]["value"]
                    ),
                    "info",
                )

            self.log_runtime_complete("recipe ended", "info")

            self.info(
                "output {} size: {:.3f}".format(output_file, out.getsize())
            )

        self.announce("ended")

    def prepare_environment(
        self, out: Output
    ) -> Tuple[OrderedDict, OrderedDict]:
        """
        Il prepores the input data to build the instrument focal planes

        Parameters
        ----------
        out: :class:`~exosim.output.output.OutputGroup`
            output group

        Returns
        -------
        dict
            soureces dict
        `~collections.OrderedDict`
            common path dictionary
        """
        out_sky = out.create_group("sky")

        sources, for_contrib = {}, {}
        parsePath = parse.ParsePath()

        if "sky" in self.mainConfig.keys():
            sky = self.mainConfig["sky"]

            # source preparation
            if "source" in sky.keys():
                parseSources = parse.ParseSources()
                sources = parseSources(
                    parameters=sky["source"],
                    wavelength=self.wl_grid,
                    time=self.time_grid,
                    output=out_sky,
                )

            # foreground preparation
            if "foregrounds" in sky.keys():
                for_contrib = parsePath(
                    parameters=sky["foregrounds"],
                    wavelength=self.wl_grid,
                    time=self.time_grid,
                    output=out_sky,
                    group_name="foregrounds",
                )

        # common optics preparation
        if "optical_path" in self.payloadConfig["Telescope"].keys():
            common_path = parsePath(
                parameters=self.payloadConfig["Telescope"]["optical_path"],
                wavelength=self.wl_grid,
                time=self.time_grid,
                output=out,
                light_path=for_contrib,
                group_name="telescope",
            )
        else:
            common_path = for_contrib
        return sources, common_path

    def run_channel(
        self,
        description: dict,
        common_path: OrderedDict,
        sources: dict,
        pointing: Tuple[u.Quantity, u.Quantity] = None,
        out: OutputType = None,
    ) -> None:
        """
        It instantiates and runs the :class:`~exosim.models.channel.Channel` for the indicated channel`

        Parameters
        ----------
        description: dict
            channel description
        common_path: `~collections.OrderedDict`
            dictionary of contributes
        sources:  dict
            dictionary containing :class:`~exosim.models.signal.Sed`
        pointing: (:class:`astropy.units.Quantity`, :class:`astropy.units.Quantity`) (optional)
            telescope pointing direction, expressed ad a tuple of RA and DEC in degrees. Default is ``None``
        out: :class:`~exosim.output.output.OutputGroup (optional)`
            output group

        """
        channel = Channel(
            parameters=description,
            wavelength=self.wl_grid,
            time=self.time_grid,
            output=out,
        )
        channel.parse_path(light_path=common_path)
        channel.estimate_responsivity()
        channel.propagate_foreground()
        channel.propagate_sources(
            sources=sources, Atel=self.payloadConfig["Telescope"]["Atel"]
        )
        channel.create_focal_planes()
        channel.rescale_contributions()
        channel.populate_focal_plane(pointing)
        # TODO test if no other stars are present
        channel.populate_bkg_focal_plane(pointing)
        if "irf_task" in description["detector"].keys():
            channel.apply_irf()

        channel.populate_foreground_focal_plane()

        channel.focal_plane.write()
        channel.frg_focal_plane.write()
        if channel.bkg_focal_plane:
            channel.bkg_focal_plane.write()

        for key, value in channel.frg_sub_focal_planes.items():
            value.write()



import operator
from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit
from scipy.signal import argrelextrema

from .pixelsNonLinearity import PixelsNonLinearity


class PixelsNonLinearityFromCorrection(PixelsNonLinearity):
    """
    This tools helps the user to find the pixel non-linearity coefficients to as inputs for ExoSim,
    starting from the measurable pixel non-linearity correction.

    In fact, the detector non linearity model, is usually written as polynomial such as

    .. math::
        Q_{det} = Q \\bigtriangleup (1 + \\sum_i a_i \\cdot Q^i)

    where :math:`Q_{det}` is the charge read by the detector, and :math:`Q` is the ideal count,
    as :math:`Q = \\phi_t`, with :math:`\\phi` being the number of electrons generated and :math:`t` being the elapsed time.
    In the equation above, :math:`\\bigtriangleup` is the operator used to defined the relation between :math:`Q_{det}` and :math:`Q`,
    which depends on the definition of the coefficients :math:`a_i` (see also equation below).

    However, it is usually the inverse operation that is known, as it's coefficients are measurable empirically:

    .. math::
        Q ={Q_{det}}\\bigtriangledown ( b_1 + \\sum_{i=2} b_i \\cdot Q_{det}^i)

    Where :math:`\\bigtriangledown` is the inverse operator of :math:`\\bigtriangleup`.
    Depending on the way the non linearity is estimated, the operator can either be a division (:math:`\\div`)
    or a multiplication (:math:`\\times`). If not specified, a division is assumed.

    The :math:`b_i` correction coefficients should be listed in the configuration file using the `pnl_coeff` keyword
    in increasing alphabetical order: `pnl_coeff_a` for :math:`b_1`,
    `pnl_coeff_b` for :math:`b_2`, `pnl_coeff_c` for :math:`b_3`,
    `pnl_coeff_d` for :math:`b_4`, `pnl_coeff_e` for :math:`b_5` and so on.
    The user can list any number of correction coefficients, and they will be automatically parsed.
    Please, note that using this notation, :math:`b_1` is not forced to be the unity.

    This class will restrieve the :math:`a_i` coefficients, starting from the the indicated :math:`b_i`.
    The results are the coefficients for a 4-th order polynomial:

    .. math::
        Q_{det} = Q \\cdot (a_1 + a_2 \\cdot Q + a_3 \\cdot Q^2 + a_4 \\cdot Q^3 + a_5 \\cdot Q^4)

    However, each pixel is different, and therefore, this class also produces a map of the coefficient for each pixel.
    Each coefficient is normally distributed around the mean value, with a standard deviation indicated in the configuration.
    If no standard deviation is indicated, the coefficients are assumed to be constant.

    The code output is a map of :math:`a_i` coefficients for each pixel, which can be injected into  :class:`~exosim.tasks.detector.applyPixelsNonLinearity.ApplyPixelsNonLinearity`.

    Attributes
    -----------
    operator_dict: dict
        dictionary of operators to use to estimates :math:`Q_{det}`
    Npt: int
        number of points used to estimate the coefficients

    Examples
    ----------

    >>> import exosim.tools as tools
    >>>
    >>> results = tools.PixelsNonLinearityFromCorrection(options_file='tools_input_example.xml',
    >>>                                                 output='output_pnl_map.h5')
    """

    operator_dict = {
        "/": operator.truediv,
        "*": operator.mul,
    }
    Npt = 2**10

    def compute_coefficients(
        self, parameters: dict, show_results: bool = True
    ) -> Tuple[List[float], float, float]:
        """
        It computes the non linearity coefficients.

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        show_results: bool
            it tells the code if showing the results in a plot. Default is `True`.
        """

        # retrieve correction coefficients
        keys = [
            k
            for k in parameters["detector"].keys()
            if "pnl_coeff" in k and "std" not in k
        ]
        if not keys:
            self.error("pixel non linear coefficients missing")
            raise KeyError("pixel non linear coefficients missing")

        keys.sort(reverse=True)
        coeff = [parameters["detector"][k] for k in keys]
        poli = np.poly1d(coeff)
        self.debug("correction coefficients: {}".format(coeff[::-1]))

        def poli_model(q, b, c, d, e):
            return 1 + b * q + c * q**2 + d * q**3 + e * q**4

        # build adu grid
        f = 1.0
        idx = [1023]
        while idx == [1023]:
            Qmax = parameters["detector"]["well_depth"]
            f += 0.1

            Qdet = np.linspace(
                1, f * Qmax.value, self.Npt
            )  # measured detector pixel counts in adu

            oper = (
                parameters["detector"]["pnl_correction_operator"]
                if "pnl_correction_operator" in parameters["detector"].keys()
                else "/"
            )
            Q = self.operator_dict[oper](
                Qdet, poli(Qdet)
            )  # linear detector pixel count in adu

            pnl_coeff, pcov = curve_fit(poli_model, Q, Qdet / Q)
            idx = np.argmin(np.abs(poli_model(Q, *pnl_coeff) - 0.95))
            self.debug("fitted coefficients: {}".format(pnl_coeff))
            self.debug("Well depth: {:.0f} (adu)".format(Q[idx]))

        if show_results:
            self._print_results(
                pnl_coeff,
                Q[idx],
            )
            self._plot(poli, Q[idx], pnl_coeff, poli_model, oper)

        return np.insert(pnl_coeff, 0, 1), Q[idx]

    def _plot(self, poli, Qmax, pnl_coeff, poli_model, oper) -> None:
        Qdet = np.linspace(
            1, Qmax * 1.2, self.Npt
        )  # detector pixel counts in adu
        Q = self.operator_dict[oper](Qdet, poli(Qdet))
        idx = np.argmin(np.abs(poli_model(Q, *pnl_coeff) - 0.95))

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        fig.suptitle("pixel linearity model")
        ax.plot(Q, Q, "k", ls=":", label="Linear pixel count")
        ax.plot(Q, Qdet, "g", label="Detector pixel count")
        ax.plot(
            Q,
            Q * poli_model(Q, *pnl_coeff),
            "r",
            label="Non-linearity model",
            ls="--",
        )
        ax.axvline(
            Qmax,
            c="b",
            label="nominal saturation: {}".format(int(np.round(Qmax))),
            ls="--",
            alpha=0.5,
        )
        ax.axvline(
            Q[idx],
            c="k",
            label="real saturation (5% from linear): {}".format(
                int(np.round(Q[idx]))
            ),
            ls=":",
            alpha=0.5,
        )

        ax.set_xlabel("$Q$ [adu]")
        ax.set_ylabel("$Q_{det}$ [adu]")
        ax.legend()
        ax.grid()
        plt.show()
        # plt.savefig('wfc3_multiply.png')



import exosim.models.signal as signal

from .task import Task


class ChainTask(Task):
    """
    Abstract class to operate on a :class:`~exosim.models.signal.Signal` and return a :class:`~exosim.models.signal.Signal`.

    Returns
    --------
    :class:`~exosim.models.signal.Radiance`
        optical element radiance
    """

    def __init__(self):
        """
        Parameters
        __________
        signal: :class:`~exosim.models.signal.Signal`
            input signal
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid
        """

        self.add_task_param("signal", "input signal")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")
        self.add_task_param("parameters", "optical element parameters dict")

    def execute(self):
        input_signal = self.get_task_param("signal")
        parameters = self.get_task_param("parameters")
        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        output_signal = self.model(
            signal=input_signal, parameters=parameters, wavelength=wl, time=tt
        )
        if issubclass(output_signal, signal.Signal):
            self.set_output(output_signal)
        else:
            self.error("output is not a Signal")
            raise TypeError("output is not a Signal")

    def model(self, parameters, wavelength, time):
        """
        Parameters
        ----------
        signal: :class:`~exosim.models.signal.Signal`
            input signal
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Signal`
            output signal c
        """
        raise NotImplementedError



from copy import deepcopy

import numpy as np

from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.iterators import iterate_over_chunks


class AddShotNoise(Task):
    """
    It adds the shot noise to the sub-exposures.
    The shot noise is added as a Poisson noise to the sub-exposures.
    If an output group is provided, it saves all the random seeds used.
    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """
        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding shot noise")
        subexposures = self.get_task_param("subexposures")
        output = self.get_task_param("output")

        random_seeds = []
        warning_raised = False
        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding shot noise"
        ):
            data = deepcopy(subexposures.dataset[chunk])
            idx = np.where(data <= 0)
            if len(idx[0]) > 0:
                data[idx] = 1e-10
                if not warning_raised:
                    self.warning(
                        "Negative and zero pixels found: values replaced with 1e-10"
                    )
                    warning_raised = True

            subexposures.dataset[chunk] = RunConfig.random_generator.poisson(
                data
            )
            subexposures.output.flush()
            random_seeds.append(RunConfig.random_seed)

        if output:
            if issubclass(output.__class__, Output):
                out_grp = output.create_group("shot noise")
                out_grp.write_list("random_seed", random_seeds)
                out_grp.write_array(
                    "chunks_index", np.arange(len(random_seeds))
                )



import gc

import astropy.units as u
import numpy as np

import exosim.tasks.instrument as instrument
from exosim.models.signal import Signal
from exosim.tasks.task import Task
from exosim.utils.klass_factory import find_task


class CreateFocalPlaneArray(Task):
    """
    It produces the focal plane array

    Returns
    -------
    :class:`~exosim.models.signal.Signal`
        focal plane array (no time evolution)
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        efficiency: :class:`~exosim.models.signal.Dimensionless`
            channel efficiency
        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("efficiency", "channel efficiency")

    def execute(self):
        parameters = self.get_task_param("parameters")

        focal_plane_dimension = (
            parameters["detector"]["spatial_pix"],
            parameters["detector"]["spectral_pix"],
        )
        self.debug("focal plane dimensions: {}".format(focal_plane_dimension))

        if "oversampling" in parameters["detector"]:
            if "irf_task" in parameters["detector"]:
                oversampling = parameters["detector"]["oversampling"]
            else:
                self.warning(
                    "The oversampling factor is set but irf_task is not provided in detector configuration. Oversampling factor is set to 1."
                )
                oversampling = 1
        else:
            oversampling = 1
        self.debug("pixel over sampling factor: {}".format(oversampling))

        focal_plane_dimension = tuple(
            [d * oversampling for d in focal_plane_dimension]
        )
        focal_plane_array = np.zeros(focal_plane_dimension)
        focal_plane_delta = parameters["detector"]["delta_pix"] / oversampling

        if parameters["type"].lower() == "spectrometer":
            # get wavelength solution
            wl_solution_task = find_task(
                parameters["wl_solution"]["wl_solution_task"],
                instrument.LoadWavelengthSolution,
            )
            wl_instance = wl_solution_task()
            wl_solution = wl_instance(parameters=parameters)

            # estimate spectral dispersion law
            spectral_pix_osr = (
                np.arange(focal_plane_array.shape[1]) * focal_plane_delta
            ).to(u.um)
            spectral_wav_osr = self._wav_osr(
                wl_solution, "spectral", parameters, spectral_pix_osr
            )

            # estimate spatial dispersion law
            spatial_pix_osr = (
                np.arange(focal_plane_array.shape[0]) * focal_plane_delta
            ).to(u.um)
            spatial_wav_osr = self._wav_osr(
                wl_solution, "spatial", parameters, spatial_pix_osr
            )

        elif parameters["type"].lower() == "photometer":
            efficiency = self.get_task_param("efficiency")

            # if we select efficiency > eff_max/e we underestimate the
            # total flux excluding to many efficiency data
            idx = np.where(
                efficiency.data[0, 0] > 0
            )  # efficiency.data[0, 0].max() / np.e)
            x_wav_osr = (
                np.linspace(
                    efficiency.spectral[idx].min().item(),
                    efficiency.spectral[idx].max().item(),
                    32 * oversampling,
                )
                * efficiency.spectral_units
            )
            # 32 is the minimum number of data we want to compute the
            # derivative. For less data points we have a wavelength
            # dependent effect if photometers efficiency

            #            x_wav_center = (efficiency.spectral[idx] *
            #                            efficiency.data[0, 0, idx]).sum() / \
            #                           efficiency.data[0, 0, idx].sum()

            spectral_wav_osr = x_wav_osr
            spatial_wav_osr = x_wav_osr

        focal_plane_array = Signal(
            spectral=spectral_wav_osr,
            spatial=spatial_wav_osr,
            data=focal_plane_array,
            metadata={
                "focal_plane_delta": focal_plane_delta,
                "oversampling": oversampling,
            },
        )
        try:
            focal_plane_array.metadata["wl_min"] = parameters["wl_min"]
            focal_plane_array.metadata["wl_max"] = parameters["wl_max"]
        except KeyError:
            pass

        self.set_output(focal_plane_array)

        del focal_plane_array
        gc.collect()

    def _wav_osr(self, wl_solution, key, parameters, pix_osr):
        if wl_solution[key].data == np.zeros_like(wl_solution["wavelength"]):
            # wavelength on each x pixel
            wav_osr = np.zeros(pix_osr.size) * u.um
        else:
            # estimate dispersion law
            par = np.polyfit(
                wl_solution[key].to(u.um).value,
                wl_solution["wavelength"].to(u.um).value,
                2,
            )
            spatial_dispersion_law = np.poly1d(par)

            # estimate center
            if any("center" in k for k in parameters["wl_solution"].keys()):
                pix_osr = self._centering(
                    parameters, wl_solution, pix_osr, key
                )

            # walength on each x pixel
            wav_osr = spatial_dispersion_law(pix_osr.to(u.um).value) * u.um
        self.debug("{} wavelength solution: {}".format(key, wav_osr))
        return wav_osr

    def _centering(
        self, parameters, wl_solution, spectral_pix_osr, key="spectral"
    ):
        """
        Shift the pixel array. If "auto" it sets the central wavelength of
        the channel in the center of the pixel array. If a wavelength  is indicated,
        it centers the wl solution on that wavelength.
        Else, it shifts the pixel array by the indicated number of pixels.

        """

        if "center" in parameters["wl_solution"].keys():
            if parameters["wl_solution"]["center"] == "auto":
                par = np.polyfit(
                    wl_solution["wavelength"].to(u.um).value,
                    wl_solution[key].to(u.um).value,
                    2,
                )
                spectral_dispersion_law_inv = np.poly1d(par)

                first_pixel = (
                    spectral_dispersion_law_inv(
                        parameters["wl_min"].to(u.um).value
                    )
                    * u.um
                )
                last_pixel = (
                    spectral_dispersion_law_inv(
                        parameters["wl_max"].to(u.um).value
                    )
                    * u.um
                )
                spectral_pix_osr_center = (first_pixel + last_pixel) / 2
                # spectral_pix_osr_center = spectral_dispersion_law_inv(
                #     central_wl.to(u.um).value) * u.um
                offset = (
                    spectral_pix_osr[0] + spectral_pix_osr[-1]
                ) / 2 - spectral_pix_osr_center
                self.debug(
                    "wl solution auto-centering mode. offset {}".format(offset)
                )
                spectral_pix_osr -= offset
            elif isinstance(parameters["wl_solution"]["center"], u.Quantity):
                par = np.polyfit(
                    wl_solution["wavelength"].to(u.um).value,
                    wl_solution[key].to(u.um).value,
                    2,
                )
                spectral_dispersion_law_inv = np.poly1d(par)
                central_pixel = (
                    spectral_dispersion_law_inv(
                        parameters["wl_solution"]["center"].to(u.um).value
                    )
                    * u.um
                )
                offset = (
                    spectral_pix_osr[0] + spectral_pix_osr[-1]
                ) / 2 - central_pixel
                self.debug(
                    "wl solution auto-centering mode. offset {}".format(offset)
                )
                spectral_pix_osr -= offset
        else:
            center = parameters["wl_solution"]["{}_center".format(key)]
            self.debug(
                "wl solution manual-centering mode. offset {}".format(center)
            )
            spectral_pix_osr -= center
        return spectral_pix_osr



from copy import deepcopy
from typing import List, Tuple

import astropy.units as u
import numpy as np
from astropy.table import Table

from exosim.models.signal import Signal
from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks
from exosim.utils.types import ArrayType


class AddCosmicRays(Task):
    """
    Task to simulate the addition of cosmic rays to sub-exposures in a detector.

    This task models the impact of cosmic rays on a detector during the exposure time. The model assumes
    that the cosmic rays can interact with the detector pixels in various predefined shapes like a cross,
    horizontal rectangle, and vertical rectangle. The number of cosmic ray events and their impact on the
    detector pixels are calculated based on the given cosmic ray flux, detector characteristics, and
    integration times for the sub-exposures.

    Notes
    -----
    - This task assumes that the cosmic ray events saturate the affected pixels, setting their value to the full well depth.
    - The probabilities for the interaction shapes are configurable. The task issues a warning if the sum of provided probabilities is not 1.
    - The cosmic ray flux is specified in ct/s/cm^2 and is scaled based on the pixel size and detector dimensions.

    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.


    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        integration_times: :class:`~astropy.units.Quantity`
            sub-exposures integration times
        outputs: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param(
            "integration_times",
            "subexposures integration times",
        )
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding cosmic rays")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        integration_times = self.get_task_param("integration_times")
        output = self.get_task_param("output")

        self.model(subexposures, parameters, integration_times, output)

    def model(
        self,
        subexposures: Signal,
        parameters: dict,
        integration_times: ArrayType,
        output=None,
    ) -> None:
        """
        Default model to simulate the addition of cosmic rays to the sub-exposures.

        This method saturates the hit pixels in the sub-exposure data based on the cosmic ray rate,
        detector properties, and the given integration times. The cosmic ray interactions could be in various shapes
        like single pixel, lines, squares, etc., which are defined in the configuration.

        Parameters
        ----------
        subexposures : Signal
            The sub-exposures' cached signal.
        parameters : dict
            The channel parameters dictionary containing detector information.
        integration_times : ArrayType
            The integration times for the sub-exposures.
        output : Output, optional
            The output file where the cosmic ray events will be stored.

        Notes
        -----
        The method assumes multiple possible shapes for the interaction:
        - Single pixel
        - Vertical line: Saturates two pixels vertically aligned.
        - Horizontal line: Saturates two pixels horizontally aligned.
        - Square: Saturates four pixels in a square.
        - Cross: Saturates five pixels in a cross shape.
        - Horizontal rectangle: Saturates six pixels in a horizontal rectangle shape.
        - Vertical rectangle: Saturates six pixels in a vertical rectangle shape.
        """
        # Extract relevant parameters from the provided dictionary
        rate = parameters["detector"]["cosmic_rays_rate"].astype(np.float64)
        rate = check_units(rate, "ct/s/cm^2")
        saturation_rate = parameters["detector"]["saturation_rate"]
        spatial_pix = parameters["detector"]["spatial_pix"]
        spectral_pix = parameters["detector"]["spectral_pix"]
        well_depth = parameters["detector"]["well_depth"]
        well_depth = check_units(well_depth, "ct", force=True).value
        pixel_size = parameters["detector"]["delta_pix"].astype(np.float64)
        pixel_size = check_units(pixel_size, "cm", force=False)
        randomise = parameters["detector"].get("cosmic_rays_randomise", False)

        # Calculate the number of events to be added to each sub-exposure
        events_counter = self.count_events(
            rate,
            pixel_size,
            spectral_pix,
            spatial_pix,
            integration_times,
            saturation_rate,
            randomise,
        )
        self.info(f"cosmic events_counter: {events_counter}")

        # Get interaction shapes and their probabilities
        shapes, probs = self.shapes_and_probs(parameters["detector"])

        # Initialize a table to store information about the cosmic ray events
        storing_info = Table(
            names=(
                "Sub-exposure",
                "spectral_pix",
                "spatial_pix",
                "saturated_pix_rand",
                "saturated_pix_seed",
                "saturated_pix",
            ),
            dtype=("i4", "i4", "i4", "f8", "i4", "U16"),
        )

        # Iterate over the sub-exposures and add cosmic ray events
        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding cosmic rays"
        ):
            data = deepcopy(subexposures.dataset[chunk])
            events_counter_chunk = events_counter[chunk[0]]
            for t in range(data.shape[0]):
                # Randomly select pixels for the cosmic ray events
                x_pix = RunConfig.random_generator.choice(
                    spectral_pix, events_counter_chunk[t]
                )
                y_pix = RunConfig.random_generator.choice(
                    spatial_pix, events_counter_chunk[t]
                )

                for x, y in zip(x_pix, y_pix):
                    info_row = [chunk[0].start + t, x, y]

                    # Randomly select a shape for the cosmic ray event
                    rand = RunConfig.random_generator.uniform(0, 1)
                    info_row.append(rand)
                    info_row.append(RunConfig.random_seed)

                    r_ = 0
                    for s, r in zip(shapes, probs):
                        r_ += r
                        if rand <= r_:
                            shape = s
                            break

                    shape_str = ""
                    for i, j in zip(shape[0], shape[1]):
                        shape_str += f"({i + x}, {j + y})"
                        try:
                            data[t, j + y, i + x] = well_depth
                        except IndexError:
                            continue

                    info_row.append(shape_str)
                    storing_info.add_row(info_row)

            # Update the sub-exposure data
            subexposures.dataset[chunk] = data
            subexposures.output.flush()

        # Write the cosmic ray events information to the output file if provided
        if output and issubclass(output.__class__, Output):
            out_grp = output.create_group("cosmic rays")
            out_grp.write_table("cosmic rays events", storing_info)
            out_grp.write_array("cosmic events counter", events_counter)
            if randomise:
                out_grp.write_scalar("random seed", RunConfig.random_seed)

    def count_events(
        self,
        rate: float,
        pixel_size: u.Quantity,
        spatial_pix: int,
        spectral_pix: int,
        integration_times: ArrayType,
        saturation_rate: float,
        randomise: bool,
    ) -> np.ndarray:
        """
        Calculate the number of cosmic ray events in each sub-exposure, randomly distributing them
        while keeping the probability proportional to the exposure time.

        Parameters
        ----------
        rate : float
            Cosmic rays flux rate in ct/s/cm^2.
        pixel_size : float
            Size of a detector pixel in cm^2.
        spatial_pix : int
            Number of spatial pixels in the detector.
        spectral_pix : int
            Number of spectral pixels in the detector.
        integration_times : :class:`~astropy.units.Quantity`
            Sub-exposure integration times.
        saturation_rate : float
            Saturation rate for the detector.
        randomise : bool
            A flag to randomize the total number of events

        Returns
        -------
        events_counter_i : np.ndarray
            An array containing the number of events for each sub-exposure, rounded to the nearest integer.

        """

        # Scale the rate based on detector characteristics
        scaled_rate = (
            rate * pixel_size * pixel_size * spatial_pix * spectral_pix
        )
        scaled_rate = check_units(scaled_rate, "ct/s")

        # Log the calculated rate
        self.info(f"cosmic rays rate: {scaled_rate}")

        # Compute the total expected number of events
        total_integration_time = np.sum(integration_times)
        self.debug(f"integration tiems: {integration_times}")

        total_events = (
            scaled_rate * total_integration_time * saturation_rate
        ).value

        if randomise:
            self.debug("Randomizing total number of events.")
            # Randomize total event
            total_events = RunConfig.random_generator.poisson(total_events)

        self.info(f"cosmic total events: {total_events}")

        # Normalize integration times to create a probability distribution
        integration_weights = integration_times / total_integration_time
        probabilities = integration_weights  # Convert to simple float array

        # Generate a random multinomial distribution of events across sub-exposures
        events_counter = np.random.multinomial(
            int(total_events), probabilities
        )

        self.debug(f"Probabilities: {probabilities}")
        self.debug(f"Probabilities sum: {np.sum(probabilities)}")
        self.debug(f"Events counter before multinomial: {int(total_events)}")

        return events_counter

    def shapes_and_probs(self, parameters: dict) -> Tuple[List, List]:
        """
        Generate cosmic ray interaction shapes and their corresponding probabilities.

        Parameters
        ----------
        parameters : dict
            A dictionary containing optional 'interaction_shapes' which is a sub-dictionary
            specifying shapes and their probabilities.

        Returns
        -------
        shapes : list of tuples
            A list of tuples representing shapes of cosmic ray interactions.
        probs : list of float
            A list of probabilities corresponding to the shapes.

        Raises
        ------
        ValueError
            If the sum of provided probabilities for all shapes is greater than 1.

        Examples
        --------
        >>> shapes, probs = AddCosmicRays.shapes_and_probs({"interaction_shapes": {"single": 0.5, "line_h": 0.3, "line_v": 0.2}})
        >>> print(shapes)
        [([0], [0]), ([0, 1], [0, 0]), ([0, 0], [0, 1]), ([0, 0, 1, 1], [0, 1, 0, 1]), ([0, 0, 0, -1, 1], [0, 1, -1, 0, 0])]
        >>> print(probs)
        [0.5, 0.3, 0.2, 0.1, 0.1]

        >>> shapes, probs = AddCosmicRays.shapes_and_probs({"interaction_shapes": {"single": 0.5, "line_h": 0.3, "line_v": 0.2}})
        >>> print(shapes)

        Warnings
        --------
        A warning is issued if the sum of provided probabilities for all shapes is not 1.
        """

        # Predefined cosmic ray interaction shapes
        predefined_shapes = {
            "single": ([0], [0]),
            "line_h": ([0, 1], [0, 0]),
            "line_v": ([0, 0], [0, 1]),
            "square": ([0, 0, 1, 1], [0, 1, 0, 1]),
            "cross": ([0, 0, 0, -1, 1], [0, 1, -1, 0, 0]),
            "rect_h": ([0, 1, 2, 0, 1, 2], [0, 0, 0, 1, 1, 1]),
            "rect_v": ([0, 0, 0, 1, 1, 1], [0, 1, 2, 0, 1, 2]),
        }

        # Initialize lists to store shapes and probabilities
        shapes, probs = [], []

        # Check if interaction shapes are defined in the parameters
        if "interaction_shapes" in parameters:
            for shape_key, prob in parameters["interaction_shapes"].items():
                if shape_key in predefined_shapes:
                    shapes.append(predefined_shapes[shape_key])
                    probs.append(prob)

            # Check if probabilities sum up to 1
            remaining_prob = 1.0 - sum(probs)
            if remaining_prob < 0:
                self.error("Total probability greater than 1.")
                raise ValueError("Total probability greater than 1.")

            # Adjust the probability for the 'single' shape if needed
            if remaining_prob != parameters["interaction_shapes"].get(
                "single", 0
            ):
                self.warning(
                    "Total probability of all interaction shapes should sum to 1. "
                    "Adjusting 'single' interaction shape to make the sum equal to 1."
                )
                shapes.append(predefined_shapes["single"])
                probs.append(remaining_prob)

            return shapes, probs

        # Default to 'single' shape with probability 1 if no shapes are defined
        return [predefined_shapes["single"]], [1.0]



import astropy.units as u
import numpy as np
from astropy.table import QTable
from scipy.interpolate import interp1d

import exosim.utils.grids as grids
from exosim.tasks.task import Task


class EstimateSpectralBinning(Task):
    """
    It computes spectral binning useful to produce the radiometric tables

    Returns
    --------
    astropy.table.QTable:
        wavelength bins table

    Raises
    --------
    TypeError:
        if the output is not :class:`~astropy.table.QTable`
    KeyError:
        if a column is missing in the output :class:`~astropy.table.QTable`

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.

    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wl_grid: :class:`~astropy.units.Quantity` (optional)
            focal plane wavelength grid. Useful for spectrometers. Default is ``None``.
        """
        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("wl_grid", "focal plane spectral grid", None)

    def execute(self):
        self.debug("spectral binning estimation started")
        parameters = self.get_task_param("parameters")
        wl_grid = self.get_task_param("wl_grid")

        table = self.model(wl_grid, parameters)

        if not isinstance(table, QTable):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        key_list = [
            "ch_name",
            "Wavelength",
            "bandwidth",
            "left_bin_edge",
            "right_bin_edge",
        ]
        for key in key_list:
            if key not in table.keys():
                self.error("missing keyword: {}".format(key))
                raise KeyError("missing keyword: {}".format(key))

        self.set_output(table)

    def model(self, wl_grid, parameters):
        """
        It runs a dedicated method if the channel is a spectrometer or a photometer.

        Parameters
        ----------
        wl_grid: :class:`~astropy.units.Quantity` (optional)
            focal plane wavelength grid. Useful for spectrometers. Default is ``None``.
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        astropy.table.QTable:
            wavelength bins table

        Raises
        --------
        AttributeError:
            if the channel type is not `spectrometer` or `photometer`
        """
        if parameters["type"].lower() == "spectrometer":
            table = self.wavelength_table_spectrometer(parameters, wl_grid)

        elif parameters["type"].lower() == "photometer":
            table = self.wavelength_table_photometer(parameters)

        else:
            self.error(
                "unsupported channel type: {}", format(parameters["type"])
            )
            raise AttributeError(
                "unsupported channel type: {}", format(parameters["type"])
            )
        return table

    def wavelength_table_spectrometer(self, description, wl_grid):
        """
        It returns the wavelength table for a spectrometer.
        The wavelength grid can be estimated in 2 modes:

        - `native` mode. If `targetR` is set to `native` the wavelength grid computed is the pixel level wavelength grid, where each bin is of the size of a pixel;
        - `fixed R` mode. If targetR` is set to a constant value, the wavelength grid is estimated using :func:`~exosim.utils.grids.wl_grid`.

        Parameters
        ----------
        description: dict
            dictionary contained the channel parameters.
        wl_grid: :class:`~astropy.units.Quantity`
            wavelength grid.

        Returns
        -------
        astropy.table.QTable:
            wavelength bins table

        Raises
        ---------
        KeyError
            Channel targetR format unsupported
        """
        table = QTable()
        # check if R is defined
        if "targetR" not in description:
            self.warning("Channel targetR missing: native R is assumed")
            description["targetR"] = "native"

        # native R
        if description["targetR"] == "native":
            wl_bin_c = wl_grid
            wl_sol = interp1d(
                wl_grid, np.arange(0, wl_grid.size), fill_value="extrapolate"
            )
            wl_bin_width = (
                wl_sol(np.arange(0, wl_grid.size) + 0.5)
                - wl_sol(np.arange(0, wl_grid.size) - 0.5)
            ) * u.um

        # fixed R
        elif isinstance(description["targetR"], int):
            wl_bin_c, wl_bin_width = grids.wl_grid(
                description["wl_min"],
                description["wl_max"],
                description["targetR"],
                return_bin_width=True,
            )
        else:
            self.error("Channel targetR format unsupported.")
            raise KeyError("Channel targetR format unsupported.")

        # preparing the table
        table["ch_name"] = [description["value"]] * wl_bin_c.size
        table["Wavelength"] = wl_bin_c
        table["bandwidth"] = wl_bin_width
        table["left_bin_edge"] = table["Wavelength"] - 0.5 * table["bandwidth"]
        table["right_bin_edge"] = (
            table["Wavelength"] + 0.5 * table["bandwidth"]
        )
        self.debug("wavelength table: \n{}".format(table))
        return table

    def wavelength_table_photometer(self, description):
        """
        It returns the wavelength table for a photometer.
        It is estimated as the central wavelength of the photometer
        with a bin width equal to the wavelength band.

        Parameters
        ----------
        description: dict
            dictionary contained the channel parameters.

        Returns
        -------
        astropy.table.QTable:
            wavelength bins table
        """
        table = QTable()

        wl_c = 0.5 * (description["wl_min"] + description["wl_max"])
        bandwidth = description["wl_max"] - description["wl_min"]
        left_bin_edge = wl_c - 0.5 * bandwidth
        right_bin_edge = wl_c + 0.5 * bandwidth
        table["ch_name"] = [description["value"]]
        table["Wavelength"] = [wl_c.value] * wl_c.unit
        table["bandwidth"] = [bandwidth.value] * bandwidth.unit
        table["left_bin_edge"] = [left_bin_edge.value] * left_bin_edge.unit
        table["right_bin_edge"] = [right_bin_edge.value] * right_bin_edge.unit
        self.debug("wavelength table: \n{}".format(table))

        return table



import numpy as np
from joblib import Parallel, delayed
from tqdm.auto import tqdm

import exosim.output as output
from exosim.tasks.task import Task
from exosim.utils import RunConfig


class ApplyIntraPixelResponseFunction(Task):
    """
    It applies the intra pixel response function to the plane

    Returns
    -------
    :class:`~exosim.models.signal.Signal`
        focal plane array (irf applied)
    """

    def __init__(self):
        """
        Parameters
        __________
        focal_plane: :class:`~exosim.models.signal.Signal`
            focal plane array
        irf_kernel	: 2D array
            the irf kernel image
        irf_kernel_delta  : scalar
            the irf kernel sampling interval in microns
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        convolution_method: str (optional)
            convolution method to use. Supported methods are:
            `fftconvolve` (:func:`scipy.signal.fftconvolve`),
            `convolve` (:func:`scipy.signal.convolve`),
            `ndimage.convolve` (:func:`scipy.ndimage.convolve`),
            `fast_convolution` (:func:`exosim.utils.convolution.fast_convolution`)
        """

        self.add_task_param("focal_plane", "focal plane")
        self.add_task_param("irf_kernel", " the irf kernel image")
        self.add_task_param(
            "irf_kernel_delta", "the irf kernel sampling interval in microns"
        )
        self.add_task_param("output", "output file", None)
        self.add_task_param(
            "convolution_method", "convolution method", "fftconvolve"
        )

    def execute(self):
        focal_plane = self.get_task_param("focal_plane")
        irf_kernel = self.get_task_param("irf_kernel")
        irf_kernel_delta = self.get_task_param("irf_kernel_delta")
        convolution_method = self.get_task_param("convolution_method")

        self.info(
            "applying pixel response function to {}".format(
                focal_plane.dataset_name
            )
        )

        output_file = self.get_task_param("output")
        if output_file and issubclass(output_file.__class__, output.Output):
            output_group = output_file.create_group("irf")
            output_group.write_array("irf_kernel", irf_kernel)
            output_group.write_array("irf_kernel_delta", irf_kernel_delta)

        img_delta = focal_plane.metadata["focal_plane_delta"]
        convolution_func, kwargs = self.select_convoltion_func(
            convolution_method, irf_kernel, irf_kernel_delta, img_delta
        )

        Parallel(n_jobs=RunConfig.n_job, require="sharedmem")(
            delayed(self._apply_to_time_slice)(
                focal_plane, t, convolution_func, kwargs
            )
            for t in tqdm(
                range(focal_plane.time.size),
                total=focal_plane.time.size,
                desc="applying {}".format(convolution_method),
            )
        )

        self.debug("looking for negative values")
        focal_plane.data = np.where(
            focal_plane.data >= 0.0, focal_plane.data, 1e-30
        )  # remove negative values

        self.set_output(focal_plane)

    def select_convoltion_func(self, method, irf, irf_delta, img_delta):
        func, kwargs = None, None

        if method == "fftconvolve":
            from scipy.signal import fftconvolve

            kwargs = {"in2": irf, "mode": "same"}
            func = fftconvolve
        elif method == "convolve":
            from scipy.signal import convolve

            kwargs = {"in2": irf, "mode": "same"}
            func = convolve
        elif method == "ndimage.convolve":
            # TODO: test this
            # TODO document this
            from scipy.ndimage import convolve

            kwargs = {"weights": irf, "mode": "wrap"}
            func = convolve
        elif method == "fast_convolution":
            from exosim.utils.convolution import fast_convolution

            kwargs = {
                "delta_im": img_delta,
                "ker": irf,
                "delta_ker": irf_delta,
            }
            func = fast_convolution
        else:
            raise ValueError(
                "{} convolution method not supported".format(method)
            )
        self.debug(
            "convolution method: {}. Selected function is {}".format(
                method, func
            )
        )
        self.debug("keywords: {}".format(kwargs))
        return func, kwargs

    # def _apply_to_time_slice(self, focal_plane, t, convolution_func, kwargs):
    #     focal_plane.data[t] = convolution_func(focal_plane.data[t], **kwargs)

    def _apply_to_time_slice(self, focal_plane, t, convolution_func, kwargs):
        """
        Applies convolution to a single time slice using padding to reduce border effects.
        """
        focal_plane.data[t] = convolution_func(focal_plane.data[t], **kwargs)



import logging
from typing import Tuple

import numpy as np
from scipy.interpolate import interp1d

from exosim.models.signal import Signal

logger = logging.getLogger(__name__)


def locate_wavelength_windows(
    psf: np.array, focal_plane: Signal, parameters: dict
) -> Tuple[np.array, np.array]:
    focal_plane_shape = (
        focal_plane.data.shape if not focal_plane.cached else focal_plane.shape
    )
    if psf.ndim == 4:
        logger.debug("PSF is 4D, will use spectral and spatial dimensions")
        return _locate_spectral_and_spatial(
            psf, focal_plane, parameters, focal_plane_shape
        )

    if psf.ndim == 3:
        logger.debug("PSF is 3D, will use spectral dimension only")
        if parameters["type"].lower() == "spectrometer":
            j0_ = np.round(
                np.arange(focal_plane_shape[2]) - psf.shape[2] // 2
            ).astype(int)
        if parameters["type"].lower() == "photometer":
            j0_ = np.repeat(
                focal_plane_shape[2] // 2 - psf.shape[2] // 2 - 1,
                focal_plane.spectral.size,
            )

        return None, j0_


def _locate_spectral_and_spatial(
    psf: np.array,
    focal_plane: Signal,
    parameters: dict,
    focal_plane_shape: tuple,
) -> Tuple[np.array, np.array]:
    if parameters["type"].lower() == "spectrometer":
        j0_ = np.round(
            np.arange(focal_plane_shape[2]) - psf.shape[3] // 2
        ).astype(int)

        if focal_plane.spatial.data == np.zeros_like(focal_plane.spatial):
            # crop PSF no spatial direction of too big
            if psf.shape[2] > focal_plane_shape[1]:
                center = psf.shape[2] // 2
                size = focal_plane_shape[1] // 2
                psf = psf[:, :, center - size : center + size, :]

            i0_ = np.repeat(
                focal_plane_shape[1] // 2 - psf.shape[2] // 2,
                focal_plane.spectral.size,
            )
        else:
            spatial_wl_sol = interp1d(
                focal_plane.spatial,
                np.arange(0, focal_plane_shape[1]),
                fill_value="extrapolate",
            )
            i0_ = np.round(
                spatial_wl_sol(focal_plane.spectral) - psf.shape[2] // 2
            ).astype(int)

    if parameters["type"].lower() == "photometer":
        j0_ = np.repeat(
            focal_plane_shape[2] // 2 - psf.shape[3] // 2 - 1,
            focal_plane.spectral.size,
        )
        i0_ = np.repeat(
            focal_plane_shape[1] // 2 - psf.shape[2] // 2 - 1,
            focal_plane.spectral.size,
        )
    return i0_, j0_



from .aperturePhotometry import AperturePhotometry
from .computePhotonNoise import ComputePhotonNoise
from .computeSignalsChannel import ComputeSignalsChannel
from .computeSubFrgSignalsChannel import ComputeSubFrgSignalsChannel
from .computeTotalNoise import ComputeTotalNoise
from .estimateApertures import EstimateApertures
from .estimateSpectralBinning import EstimateSpectralBinning
from .multiaccum import Multiaccum
from .saturationChannel import SaturationChannel



import astropy.units as u
from astropy.coordinates import SkyCoord

from exosim.tasks.task import Task


class ComputeSourcesPointingOffset(Task):
    """
    It computes the source offset on the focal plane respect to the pointing direction.
    The offset is in units of subpixels.

    Returns
    -------
     int
       offset in the spatial direction
     int
       offset in the spectral direction

    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            channel parameter dictionary. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        source:  dict
            dictionary containing :class:`~exosim.models.signal.Sed` metadata
        pointing: tuple
            telescope pointing direction, expressed as a tuple of RA and DEC. Default is ``None``

        """

        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("source", "source source description dictionary")
        self.add_task_param("pointing", "telescope pointing")

    def execute(self):
        parameters = self.get_task_param("parameters")
        source = self.get_task_param("source")
        pointing = self.get_task_param("pointing")

        compute = True if pointing else False
        if "ra" in source["parsed_parameters"].keys():
            self.debug("RA found in source description")
            compute *= True
        else:
            compute *= False
        if "dec" in source["parsed_parameters"].keys():
            self.debug("DEC found in source description")
            compute *= True
        else:
            compute *= False
        if "plate_scale" in parameters["detector"].keys():
            self.debug("plate scale found in source description")
            compute *= True
        else:
            compute *= False

        if compute:
            aov_spatial, aov_spectral = angle_of_view(
                parameters["detector"]["plate_scale"],
                parameters["detector"]["delta_pix"],
                parameters["detector"]["oversampling"],
            )
            self.debug(
                "Angle of View estimated: {}, {}".format(
                    aov_spatial, aov_spectral
                )
            )
            c_tel = SkyCoord(pointing[0], pointing[1], frame="icrs")
            c_source = SkyCoord(
                source["parsed_parameters"]["ra"],
                source["parsed_parameters"]["dec"],
                frame="icrs",
            )
            offset_spatial = (
                c_tel.ra.deg - c_source.ra.deg
            ) / aov_spatial.value
            offset_spectral = (
                c_tel.dec.deg - c_source.dec.deg
            ) / aov_spectral.value
            self.debug(
                "offset estimated:{} {}".format(
                    offset_spectral, offset_spatial
                )
            )

        else:
            self.debug(
                "Angle of View computation skipped: missing information"
            )
            offset_spectral = offset_spatial = 0

        self.set_output([int(offset_spatial), int(offset_spectral)])


def angle_of_view(plate_scale, delta_pix, ovs):
    """
    Computes the Angle of View for a single pixel

    Parameters
    ----------
    plate_scale: :class:`astropy.units.Quantity`
        plate scale
    delta_pix: :class:`astropy.units.Quantity`
        size of a pixel
    Returns
    -------
    :class:`astropy.units.Quantity`
        angle of view in deg of each subpixel in the spatial direction
    :class:`astropy.units.Quantity`
        angle of view in deg of each subpixel in the spectral direction
    """

    def _compute_angle(plate_scale):
        if isinstance(plate_scale, u.Quantity):
            try:
                plate_scale.to(u.deg / u.micron)
                angle = plate_scale * delta_pix / ovs
            except u.UnitConversionError:
                try:
                    plate_scale.to(u.arcsec / u.pixel)
                    angle = plate_scale * u.pixel / ovs
                except u.UnitConversionError:
                    raise u.UnitConversionError(
                        "wrong plate scale units: {}".format(plate_scale.unit)
                    )

        else:
            raise OSError("missing plate scale units")

        return angle.to(u.deg)

    if isinstance(plate_scale, dict):
        spatial_angle = _compute_angle(plate_scale["spatial"])
        spectral_angle = _compute_angle(plate_scale["spectral"])
    else:
        spatial_angle = spectral_angle = _compute_angle(plate_scale)

    return spatial_angle, spectral_angle



import os
import xml.etree.ElementTree as ET
from collections import OrderedDict

from astropy import units as u
from astropy.table import Table

from exosim.tasks.task import Task
from exosim.utils.runConfig import RunConfig


class LoadOptions(Task):
    """
    Reads the XML file with payload parameters and returns an object with attributes related to the input data.

    Attributes
    ----------
    config_path : str
        Configuration path.

    Returns
    -------
    dict
        Parsed XML input file.

    Raises
    ------
    IOError
        If the indicated file is not found or the format is not supported.

    Examples
    --------
    >>> load_options = LoadOptions()
    >>> options = load_options(filename='path/to/file.xml')
    """

    def __init__(self):
        super().__init__()  # Explicitly call the parent class constructor
        self.add_task_param("filename", "Input option file name")
        self.add_task_param(
            "config_path", "On-run setting for ConfigPath", None
        )

    def execute(self):
        self._filename = self.get_task_param("filename")
        self.config_path = (
            self.get_task_param("config_path") or "__ConfigPath__"
        )

        RunConfig.config_file_list.append(os.path.abspath(self._filename))

        self._check_format()
        root = self._get_root()
        options_dict = self._parse_xml(root)
        self._finalise(options_dict)
        options_dict = self._substitute_keywords(options_dict)
        options_dict = _clean_dict(options_dict)
        self.set_output(options_dict)

    def _check_format(self):
        if not self._filename.endswith(".xml"):
            self.error("Wrong input format: Expected an XML file.")
            raise OSError(
                "Unsupported file format. Please provide an XML file."
            )

    def _get_root(self):
        try:
            self.debug(f"Input option file found: {self._filename}")
            return ET.parse(self._filename).getroot()
        except FileNotFoundError:
            self.error(f"No input option file found: {self._filename}")
            raise FileNotFoundError(f"Input file not found: {self._filename}")

    def _parse_xml(self, root):
        root_dict = {}
        for child in root:
            child_dict = self._parse_xml(child)
            child_dict.update(child.attrib)

            value = self._compact_string(child.text)
            if value:
                value = self._convert_value(value)
                if "unit" in child_dict:
                    unit_name = child_dict.pop("unit")
                    if unit_name == "dimensionless":
                        unit_name = ""
                    value = value * u.Unit(unit_name)
                if isinstance(value, str) and "__ConfigPath__" in value:
                    value = value.replace("__ConfigPath__", self.config_path)
                child_dict["value"] = value

            if child.tag == "ConfigPath":
                self.config_path = value

            if child.tag in root_dict:
                existing_attr = root_dict[child.tag]
                if isinstance(existing_attr, OrderedDict):
                    existing_attr[value] = child_dict
                else:
                    dtmp = OrderedDict(
                        [
                            (existing_attr.get("value"), existing_attr),
                            (value, child_dict),
                        ]
                    )
                    root_dict[child.tag] = dtmp
            else:
                root_dict[child.tag] = child_dict

        if "datafile" in root_dict:
            datafile = root_dict["datafile"]["value"].replace(
                "__ConfigPath__", self.config_path
            )
            if not os.path.exists(datafile):
                self.error(f"Datafile not found: {datafile}")
                raise FileNotFoundError(f"Datafile not found: {datafile}")
            try:
                root_dict["data"] = self._read_data_table(datafile)
            except OSError:
                self.error(f"Cannot read input file: {datafile}")
                raise OSError("Error reading the input data file.")

        return root_dict

    def _substitute_keywords(self, root_dict):
        """
        Substitutes keywords in the root_dict with their corresponding values.

        Parameters
        ----------
        root_dict : dict
            The dictionary parsed from the XML.

        Returns
        -------
        dict
            Updated dictionary with substituted keywords.
        """
        # Extract keywords (elements starting with '__')
        keywords = {
            key.strip("_"): value["value"]
            for key, value in root_dict.items()
            if isinstance(value, dict)
            and "value" in value
            and key.startswith("__")
        }

        def substitute_value(value):
            if isinstance(value, str):
                for keyword, replacement in keywords.items():
                    value = value.replace(f"__{keyword}__", replacement)
            return value

        def recursive_substitute(d):
            for key, value in d.items():
                if isinstance(value, dict):
                    recursive_substitute(value)
                elif isinstance(value, str):
                    d[key] = substitute_value(value)

        recursive_substitute(root_dict)
        return root_dict

    def _finalise(self, dictionary):
        xml_entry = dictionary.pop("config", None)
        if xml_entry:
            xmlfile = xml_entry["value"].replace(
                "__ConfigPath__", self.config_path
            )
            xmlfile = os.path.expanduser(xmlfile)
            if not os.path.exists(xmlfile):
                self.error(f"Referenced config file not found: {xmlfile}")
                raise FileNotFoundError(
                    f"Referenced config file not found: {xmlfile}"
                )
            sub_system_dict = LoadOptions()(
                filename=xmlfile, config_path=self.config_path
            )
            dictionary.update(sub_system_dict)

        for key, item in dictionary.items():
            if isinstance(item, dict):
                self._finalise(item)

    def _compact_string(self, string):
        return string.replace("\n", "").strip() if string else ""

    def _convert_value(self, value):
        try:
            return int(value)
        except ValueError:
            try:
                return float(value)
            except ValueError:
                return self._convert_boolean(value)

    def _convert_boolean(self, value):
        if value == "True":
            return True
        elif value == "False":
            return False
        return value

    def _read_data_table(self, datafile):
        data_type = os.path.splitext(datafile)[1]
        try:
            data = Table.read(
                os.path.expanduser(datafile),
                fill_values=[("#N/A", "0"), ("N/A", "0"), ("", "0")],
                format="ascii" + data_type,
            )
        except Exception as exc:
            raise Exception(f"{datafile} caused the exception") from exc

        for col in data.columns:
            if hasattr(data[col], "fill_value"):
                data[col].fill_value = 0.0

        return data


def _clean_dict(input_dict):
    """
    Cleans an input dictionary by removing the "value" notation and comments.
    It can be applied recursively.

    Parameters
    ----------
    input_dict : dict
        The dictionary to clean.

    Returns
    -------
    dict
        Cleaned dictionary.
    """
    input_dict.pop("comment", None)

    for key in list(input_dict.keys()):
        if isinstance(input_dict[key], dict):
            keys_list = list(input_dict[key].keys())
            if keys_list == ["unit", "value"]:
                input_dict[key] = input_dict[key]["value"]
            elif keys_list == ["value"]:
                input_dict[key] = input_dict[key]["value"]
            else:
                _clean_dict(input_dict[key])

    return input_dict



# This code is inspired by the code devolped for TauREx 3.1.
# Therefore, we attach here TauREx3.1 license:
#
# BSD 3-Clause License
#
# Copyright (c) 2019, Ahmed F. Al-Refaie, Quentin Changeat, Ingo Waldmann, Giovanna Tinetti
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from abc import ABC

import numpy as np

import exosim.log as log


class Output(log.Logger):
    def __init__(self):
        super().__init__()

    def open(self):
        """It opens the output file"""
        raise NotImplementedError

    def create_group(self, group_name):
        """
        it creates an :class:`~exosim.output.output.OutputGroup`.

        Parameters
        ----------
        group_name: str
            group name

        Returns
        -------
        OutputGroup

        """
        raise NotImplementedError

    def close(self):
        """It closes the output file"""
        raise NotImplementedError

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, type, value, tb):
        self.close()

    def store_dictionary(self, dictionary, group_name=None):
        """
        it stores a full dictionary inside an :class:`~exosim.output.output.OutputGroup`.

        Parameters
        ----------
        dictionary: dict
            data to store
        group_name: str
            group name for the data
        """

        from exosim.output.utils import (
            recursively_save_dict_contents_to_output,
        )

        out = self
        if group_name is not None:
            out = self.create_group(group_name)

        recursively_save_dict_contents_to_output(out, dictionary)

    def getsize(self):
        """It returns the output file size"""
        raise NotImplementedError

    def delete_data(self, key):
        raise NotImplementedError


class OutputGroup(Output, ABC):
    def __init__(self, name):
        super().__init__(name)
        self._name = name

    def write_array(self, array_name, array, metadata=None):
        """
        Method to store :class:`~numpy.ndarray`.

        Parameters
        ----------
        array_name: str
            dataset name
        array: :class:`~numpy.ndarray`
            data to store
        metadata: dict
            metadata to attach
        """
        raise NotImplementedError

    def write_list(self, list_name, list_array, metadata=None):
        """
        Method to store lists.

        Parameters
        ----------
        list_name: str
            dataset name
        list_array: list
            data to store
        metadata: dict
            metadata to attach
        """
        arr = np.array(list_array)
        self.write_array(list_name, arr)

    def write_scalar(self, scalar_name, scalar, metadata=None):
        """
        Method to store scalars.

        Parameters
        ----------
        scalar_name: str
            dataset name
        scalar: int or float
            data to store
        metadata: dict
            metadata to attach
        """
        raise NotImplementedError

    def write_string(self, string_name, string, metadata=None, replace=False):
        """
        Method to store strings.

        Parameters
        ----------
        string_name: str
            dataset name
        string: str
            data to store
        metadata: dict
            metadata to attach
        """
        raise NotImplementedError

    def write_string_array(self, string_name, string_array, metadata=None):
        """
        Method to store :class:`~numpy.ndarray` of strings.

        Parameters
        ----------
        string_name: str
            dataset name
        string_array: :class:`~numpy.ndarray`
            data to store
        metadata: dict
            metadata to attach
        """
        raise NotImplementedError

    def write_table(self, table_name, table, metadata=None, replace=False):
        """
        Method to store :class:`~astropy.table.Table` or :class:`~astropy.table.QTable`.

        Parameters
        ----------
        table_name: str
            dataset name
        table: :class:`~astropy.table.Table` or :class:`~astropy.table.QTable`
            data to store
        metadata: dict
            metadata to attach
        replace: bool
            if ``True``, it replaces the table in the outpur file if already existing.
            Default is ``False``
        """
        raise NotImplementedError

    def write_quantity(self, quantity_name, quantity, metadata=None):
        """
        Method to store :class:`~astropy.units.Quantity`.

        Parameters
        ----------
        quantity_name: str
            dataset name
        quantity: :class:`~astropy.units.Quantity`
            data to store
        metadata: dict
            metadata to attach
        """
        raise NotImplementedError

    def create_group(self, group_name):
        """
        it creates an :class:`OutputGroup` as a subgroup

        Parameters
        ----------
        group_name: str
            group name

        Returns
        -------
        OutputGroup

        """
        raise NotImplementedError

    def delete_data(self, key):
        """
        It deletes a specific key and its associated data from the data store.

        Parameters
        ----------
        key : str
            The key identifying the data to be deleted.

        Raises
        ------
        KeyError
            If the key is not found in the data store
        """
        raise NotImplementedError



import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

import exosim.models.signal as signal
import exosim.utils as utils
from exosim.output import SetOutput
from exosim.utils import RunConfig, check_units

from .exosimTool import ExoSimTool


class QuantumEfficiencyMap(ExoSimTool):
    """
    Produces the channels quantum efficiency variation map

    Returns
    --------
    dict:
        channels' quantum efficiency map

    Raises
    ------
    TypeError:
        if the output is not a :class:`~exosim.models.signal.Signal` class

    Examples
    ----------

    >>> import exosim.tools as tools
    >>>
    >>> results = tools.QuantumEfficiencyMap(options_file='tools_input_example.xml',
    >>>                                      output='output_qe_map.h5')
    """

    def __init__(self, options_file, output=None):
        """
        Parameters
        __________
        parameters: str or dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: str (optional)
            output file
        """
        super().__init__(options_file)

        self.info("creating qe map")

        time_grid = utils.grids.time_grid(
            self.options["time_grid"]["start_time"],
            self.options["time_grid"]["end_time"],
            self.options["time_grid"]["low_frequencies_resolution"],
        )
        output = SetOutput(output)

        with output.use() as out:
            for ch in self.ch_list:
                qe_map = self.model(self.ch_param[ch], time_grid)

                self.results.update({ch: qe_map})

                self.debug("qe variation map: {}".format(qe_map.data))

                qe_map.write(out, ch)

    def model(self, parameters, time):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Signal`
            channel quantum efficiency map

        """

        # variance map
        variance_map = RunConfig.random_generator.normal(
            1,
            parameters["detector"]["qe_sigma"],
            (
                1,
                parameters["detector"]["spatial_pix"],
                parameters["detector"]["spectral_pix"],
            ),
        )
        t = [0] * u.hr
        # temporal variation
        if "qe_aging_factor" in parameters["detector"].keys():
            t0 = parameters["detector"]["qe_aging_time_scale"]
            t0 = check_units(t0, u.hr, force=True)

            # creates the aging effect
            qe_aging = RunConfig.random_generator.normal(
                1,
                parameters["detector"]["qe_aging_factor"],
                (
                    1,
                    parameters["detector"]["spatial_pix"],
                    parameters["detector"]["spectral_pix"],
                ),
            )

            # remove values greater than 1, because QE should become smaller
            qe_aging[qe_aging > 1] = 2 - qe_aging[qe_aging > 1]

            # create the map evolution over time
            variance_map = np.vstack((variance_map, variance_map * qe_aging))
            t = np.hstack((t, t0))
            variance_map = variance_map.reshape(
                t.size, variance_map.shape[1] * variance_map.shape[2]
            )
            f = interp1d(t, variance_map, fill_value="extrapolate", axis=0)
            variance_map = f(time)
            variance_map = variance_map.reshape(
                time.size,
                parameters["detector"]["spatial_pix"],
                parameters["detector"]["spectral_pix"],
            )
            t = time

        qe_map = signal.Dimensionless(
            data=variance_map,
            spectral=np.arange(0, parameters["detector"]["spectral_pix"])
            * u.pix,
            time=t,
        )

        qe_map.temporal_rebin(time)

        return qe_map



from copy import deepcopy

import numpy as np

from exosim.output import Output
from exosim.tasks.task import Task
from exosim.utils import RunConfig
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks


class AddReadNoiseMapNumpy(Task):
    """
    This Task simulates the read noise as a normal distribution which parameters can be defined with a map indicated in the configuration file under `read_noise_filename` keyword.
    The input must be a NPY format file (see `numpy documentation <https://numpy.org/devdocs/reference/generated/numpy.lib.format.html>`_) containing the map of the distribution standard deviation for each pixel.

    A different realisations of the same distribution is added to each pixel of each sub-exposure.
    If an output group is provided, it saves all the random seeds used.

    """

    def __init__(self):
        """
        Parameters
        ----------
        subexposures: :class:`~exosim.models.signal.Counts`
            sub-exposures cached signal
        parameters: dict
            channel parameters dictionary
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("subexposures", "sub-exposures cached signal")
        self.add_task_param("parameters", "channel parameters dictionary")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("adding read noise")
        subexposures = self.get_task_param("subexposures")
        parameters = self.get_task_param("parameters")
        output = self.get_task_param("output")

        read_noise_file = parameters["detector"]["read_noise_filename"]
        read_noise_sigma = np.load(read_noise_file)
        read_noise_sigma = check_units(read_noise_sigma, "ct", force=True)

        if read_noise_sigma.shape != subexposures.dataset[0].shape:
            self.error(
                "wrong shape: expected {} but got {}".format(
                    subexposures.dataset[0].shape, read_noise_sigma.shape
                )
            )
            raise OSError()

        random_seeds = []

        for chunk in iterate_over_chunks(
            subexposures.dataset, desc="adding read noise"
        ):
            data = deepcopy(subexposures.dataset[chunk])
            subexposures.dataset[chunk] = (
                data
                + RunConfig.random_generator.normal(
                    0, read_noise_sigma, data.shape
                ).astype(np.float64)
            )
            subexposures.output.flush()
            random_seeds.append(RunConfig.random_seed)

        if output:
            if issubclass(output.__class__, Output):
                out_grp = output.create_group("read noise")
                out_grp.write_list("random_seed", random_seeds)
                out_grp.write_array(
                    "chunks_index", np.arange(len(random_seeds))
                )



from collections import OrderedDict
from typing import List, Union

import exosim.tasks.load as load
from exosim.log import Logger


class ExoSimTool(Logger):
    """
    *Abstract class*

    Base class for exosim tools.

    Attributes
    ----------------
    ch_param: dict
        dictionary for the channels' configurations
    options: dict
        configurations dictionary
    results: dict
        dictionary for the results output
    """

    def __init__(self, options_file: Union[str, dict]) -> None:
        super().__init__()
        if isinstance(options_file, str):
            loadOption = load.LoadOptions()
            self.options = loadOption(filename=options_file)
        elif isinstance(options_file, dict):
            self.options = options_file

        self.ch_param = {}
        if isinstance(self.options["channel"], OrderedDict):
            for key, value in self.options["channel"].items():
                self.ch_param[key] = value
        else:
            self.ch_param[self.options["channel"]["value"]] = self.options[
                "channel"
            ]

        self.results = {}

    @property
    def ch_list(self) -> List[str]:
        """list of channel names"""
        if isinstance(self.options["channel"], OrderedDict):
            ch_list = list(self.options["channel"])
            ch_list.sort()
        else:
            ch_list = [self.options["channel"]["value"]]
        return ch_list



import copy
from copy import deepcopy

import astropy.units as u
import numpy as np
from numba import jit, prange
from scipy.interpolate import RectBivariateSpline

import exosim.output as output
from exosim.models.signal import Counts
from exosim.tasks.subexposures.computeReadingScheme import ComputeReadingScheme
from exosim.tasks.subexposures.estimateChJitter import EstimateChJitter
from exosim.tasks.task import Task
from exosim.utils.checks import check_units
from exosim.utils.iterators import iterate_over_chunks, searchsorted
from exosim.utils.operations import operate_over_axis


class PrepareInstantaneousReadOut(Task):
    """
    This task prepares the instantaneous read out.
    It calls  :class:`~exosim.tasks.subexposures.computeReadingScheme.ComputeReadingScheme` to compute the ramp sampling scheme,
    and :class:`~exosim.tasks.subexposures.estimateChJitter.EstimateChJitter` to scale the input pointing jitter to the focal planet pixel units.
    The jittering is based on the focal plane oversampling factor.

    Returns
    --------
    dict
        readout_parameters dict
    :class:`~astropy.units.Quantity`
        sub-exposures integration times

    """

    def __init__(self):
        """
        Parameters
        ----------
        main_parameters: dict
            main parameters dict
        focal_plane: :class:`~exosim.models.signal.CountsPerSecond`
            channel focal plane
        pointing_jitter: (:class:`~astropy.units.Quantity`, :class:`~astropy.units.Quantity`,  :class:`~astropy.units.Quantity`)
            Tuple containing the pointing jitter in the spatial and spectral direction expressed in units of deg, adn jitter time expressed as sec.
        parameters: dict
            dictionary containing the channel parameters.
            This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: str or :class:`~exosim.output.hdf5.hdf5.HDF5Output` or :class:`~exosim.output.hdf5.hdf5.HDF5OutputGroup`
            output file
        """
        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param("main_parameters", "channel parameters dict")
        self.add_task_param("focal_plane", "loaded focal plane")
        self.add_task_param("pointing_jitter", "")
        self.add_task_param("output_file", "output file")
        self.store_dict = {}

    def execute(self):
        parameters = self.get_task_param("parameters")
        main_parameters = self.get_task_param("main_parameters")
        pointing_jitter = self.get_task_param("pointing_jitter")
        focal_plane = self.get_task_param("focal_plane")
        output_file = self.get_task_param("output_file")

        osf = focal_plane.metadata["oversampling"]

        if pointing_jitter != (None, None, None):
            estimateChJitter = EstimateChJitter()
            jitter_spe, jitter_spa, y_jit, x_jit, jit_time = estimateChJitter(
                pointing_jitter=pointing_jitter, parameters=parameters
            )

            new_freq = check_units(jit_time[1] - jit_time[0], "Hz")
            readout_freq = check_units(
                parameters["readout"]["readout_frequency"], "Hz"
            )
            readout_oversampling = new_freq / readout_freq
        else:
            readout_oversampling = 1

        computeReadingScheme = ComputeReadingScheme()
        (
            clock,
            base_mask,
            frame_sequence,
            number_of_exposures,
        ) = computeReadingScheme(
            parameters=parameters,
            main_parameters=main_parameters,
            readout_oversampling=readout_oversampling,
            output_file=output_file,
        )

        # Number of frames contributing to each NDR
        ndr_end_cumulative_sequence = np.ma.array(
            frame_sequence.cumsum(),
            mask=np.tile(np.logical_not(base_mask), number_of_exposures),
        ).compressed()

        base_mask_start = base_mask
        base_mask_start[0] = 1
        base_mask_start[-2] = 0

        ndr_start_cumulative_sequence = np.ma.array(
            frame_sequence.cumsum(),
            mask=np.tile(np.logical_not(base_mask_start), number_of_exposures),
        ).compressed()

        ndr_integration_times = (
            ndr_end_cumulative_sequence - ndr_start_cumulative_sequence
        ) * clock

        focal_time_sampled = (focal_plane.time * focal_plane.time_units).to(
            u.s
        )
        fp_time = searchsorted(
            focal_time_sampled, ndr_start_cumulative_sequence * clock
        )

        self.store_dict = {
            "number_of_exposures": number_of_exposures,
            "frame_sequence": frame_sequence,
            "ndr_end_cumulative_sequence": ndr_end_cumulative_sequence,
            "ndr_start_cumulative_sequence": ndr_start_cumulative_sequence,
            "ndr_integration_times": ndr_integration_times,
            "simulation_clock": clock,
            "fp_time": fp_time,
        }

        if pointing_jitter != (None, None, None):
            self.debug("Pointing jitter found")

            # check if the jitter is well sampled

            # take min rms if min is not zero
            jitter_rms_array = np.array(
                [np.sqrt(np.mean(y_jit**2)), np.sqrt(np.mean(x_jit**2))]
            )

            try:
                jitter_rms = min(
                    jitter_rms_array[np.nonzero(jitter_rms_array)]
                )
            except ValueError:
                self.debug("jitter rms is zero in both directions")
                jitter_rms = 0
            self.debug("jitter rms: {:.2f}".format(jitter_rms))

            # check if specific resolution is required else set it to 3 by default
            try:
                jitter_res = parameters["detector"][
                    "jitter_rms_min_resolution"
                ]
                self.debug("found jitter rms minimum resolution")
            except KeyError:
                jitter_res = 3.0

            # check if magnification is wanted or needed
            mag = 1
            # check if magnification is needed
            if jitter_rms != 0 and jitter_rms < jitter_res:
                self.debug("jitter rms < 3: focal plane resampling needed")
                mag = np.ceil(jitter_res / jitter_rms).astype(int)

            # check if magnification is suggested
            try:
                forced_mag = parameters["detector"]["jitter_resampler_mag"]
                self.debug("found jitter resampler magnification")
            except KeyError:
                forced_mag = None
            # if magnification is suggested choose between the suggested and the computed
            if forced_mag:
                if forced_mag < mag:
                    self.warning(
                        "suggested jitter magnification ({}) is too small: {} used instead".format(
                            forced_mag, mag
                        )
                    )
                else:
                    mag = forced_mag

            # producing diagnostic info
            spe_jit_ave = (
                self._average_pointing(
                    jitter_spe.value,
                    ndr_start_cumulative_sequence,
                    ndr_end_cumulative_sequence,
                )
                * jitter_spe.unit
            )
            spa_jit_ave = (
                self._average_pointing(
                    jitter_spa.value,
                    ndr_start_cumulative_sequence,
                    ndr_end_cumulative_sequence,
                )
                * jitter_spa.unit
            )
            x_jit_ave = (
                self._average_pointing(
                    x_jit,
                    ndr_start_cumulative_sequence,
                    ndr_end_cumulative_sequence,
                )
                / osf
                * u.pix
            )
            y_jit_ave = (
                self._average_pointing(
                    y_jit,
                    ndr_start_cumulative_sequence,
                    ndr_end_cumulative_sequence,
                )
                / osf
                * u.pix
            )
            jit_indexes = []
            for start, stop in zip(
                ndr_start_cumulative_sequence, ndr_end_cumulative_sequence
            ):
                jit_indexes.append(np.arange(start, stop).astype(int))
            self.store_dict.update(
                {
                    "y_jit": y_jit / osf * u.pix,
                    "x_jit": x_jit / osf * u.pix,
                    "y_jit_averaged": y_jit_ave,
                    "x_jit_averaged": x_jit_ave,
                    "spe_jit_averaged": spe_jit_ave,
                    "spa_jit_averaged": spa_jit_ave,
                    "jit_indexes": {"se": jit_indexes},
                    "mag": mag,
                    "effective_osf": osf * mag,
                }
            )

        for k, v in self.store_dict.items():
            self.debug("{}: {}".format(k, v))

        if issubclass(output_file.__class__, output.Output):
            output_file.store_dictionary(
                self.store_dict, "instantaneous_readout_params"
            )

        self.set_output([self.store_dict, ndr_integration_times])

    def force_power_conservation(
        self, out, parameters, focal_plane, fp_time, osf
    ):
        # to compute the total power on the focal plane I use the undersampled focal plane

        total_power = np.empty(out.dataset.shape[0])
        desired_power = np.empty(out.dataset.shape[0])

        for chunk in iterate_over_chunks(
            out.dataset,
            desc="computing median incoming power {}".format(
                parameters["value"]
            ),
        ):
            # computing the total power in the jittered focal plane
            dset = out.dataset[chunk]
            total_power[chunk[0]] = dset.sum(axis=-1).sum(axis=-1)

            # computing the desired power from the original focal planes
            fp_time_ = fp_time[chunk[0]]
            fp_times = list(set(fp_time_))
            for time_id in fp_times:
                mask = np.where(fp_time_ == time_id)[0]
                # I estimated the expected power from the oversampled focal plane
                desired_power[chunk[0]][mask] = (
                    np.sum(focal_plane.data[time_id]) / osf**2
                )

        # applying integration time to the jittered focal planes
        for chunk in iterate_over_chunks(
            out.dataset,
            desc="forcing conservation of power {}".format(
                parameters["value"]
            ),
        ):
            dset = out.dataset[chunk]
            dset = operate_over_axis(
                dset, desired_power[chunk[0]] / total_power[chunk[0]], 0, "*"
            )

            out.dataset[chunk] = dset

            out.output.flush()

        self.store_dict.update(
            {"median_power": desired_power, "total_power": total_power}
        )

    @staticmethod
    @jit(nopython=True, parallel=True)
    def _average_pointing(
        jitter: np.ndarray, start: np.ndarray, stop: np.ndarray
    ) -> np.ndarray:
        jitter = jitter.astype(np.float64)
        start = start.astype(np.int64)
        stop = stop.astype(np.int64)

        out = np.empty(start.shape[0], dtype=np.float64)
        for i in prange(start.shape[0]):
            s = start[i]
            e = stop[i]
            acc = 0.0
            for j in range(s, e):
                acc += jitter[j]
            out[i] = acc / (e - s)
        return out



import astropy.units as u
from astropy.table import QTable

from exosim.tasks.task import Task

from .aperturePhotometry import AperturePhotometry


class ComputeSubFrgSignalsChannel(Task):
    """
    It iteratively estimates the radiometric signals on the foregrounds sub focal planes for a channel
    and returns a table with all the contributions.

    Returns
    --------
    :class:`astropy.table.QTable`
        signal table

    Raises
    --------
    TypeError:
        if the output is not :class:`~astropy.table.QTable`

    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        ----------
        table: :class:`astropy.table.QTable`
            apertures table
        ch_name: str
            channel name
        input_file: :class:`~exosim.output.output.Output`
            input HDF5 file
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """
        self.add_task_param("table", "channel table")
        self.add_task_param("ch_name", "channel name")
        self.add_task_param(
            "input_file", "input file containing the focal planes"
        )
        self.add_task_param("parameters", "channel parameters dict", None)

    def execute(self):
        table = self.get_task_param("table")
        ch = self.get_task_param("ch_name")
        input_file = self.get_task_param("input_file")
        parameters = self.get_task_param("parameters")

        self.debug("Computing foreground signals table for {}".format(ch))
        new_table = self.model(ch, table, input_file, parameters)

        if not isinstance(new_table, QTable):
            self.error("wrong output format")
            raise TypeError("wrong output format")

        self.set_output(new_table)

    def model(self, ch, table, input_file, parameters):
        """
        It iteratively estimates the radiometric signals on the foregrounds sub focal plane for a channel
        and returns a table with all the contributions.
        It uses :func:`photutils.aperture.aperture_photometry` with the apertures
        from :class:`~exosim.tasks.radiometric.estimateApertures.EstimateApertures`.

        Parameters
        ----------
        table: :class:`astropy.table.QTable`
            apertures table
        ch_name: str
            channel name
        input_file: :class:`~exosim.output.output.Output`
            input HDF5 file
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        :class:`astropy.table.QTable`
            signal table
        """

        new_table = QTable()
        with input_file.open() as f:
            if "sub_focal_planes" in f["channels"][ch].keys():
                for frg in f["channels"][ch]["sub_focal_planes"].keys():
                    # prepare the focal plane
                    self.debug(
                        "extracting {} focal plane. Oversampling removed".format(
                            frg
                        )
                    )
                    sub_f = f["channels"][ch]["sub_focal_planes"][frg]
                    osf = sub_f["metadata"]["oversampling"][()]
                    focal_plane_units = u.Unit(sub_f["data_units"][()])
                    focal_plane = (
                        sub_f["data"][0, osf // 2 :: osf, osf // 2 :: osf]
                        * focal_plane_units
                    )

                    aperturePhotometry = AperturePhotometry()

                    self.debug(
                        "aperture photometry for {} focal plane".format(frg)
                    )
                    signal_in_phot = aperturePhotometry(
                        table=table, focal_plane=focal_plane
                    )

                    name = frg.split("_")[-1]
                    new_table["{}_signal_in_aperture".format(name)] = (
                        signal_in_phot * focal_plane_units
                    )
        return new_table



import os

import numpy as np
from astropy.io import ascii
from astropy.table import Table

from exosim.utils import RunConfig

from .exosimTool import ExoSimTool


class DeadPixelsMap(ExoSimTool):
    """
    Produces the channel dead pixel map

    Returns
    --------
    dict
        channels' dead pixels maps

    Raises
    ------
    TypeError:
        if the output is not a :class:`~astropy.table.Table` class

    Examples
    ----------

    >>> import exosim.tools as tools
    >>>
    >>> tools.DeadPixelsMap(options_file='tools_input_example.xml',
    >>>                     output='./')
    """

    def __init__(self, options_file, output=None):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        output: str (optional)
            output directory
        """
        super().__init__(options_file)

        self.info("creating dead pixel map")

        for ch in self.ch_list:
            dp_map = self.model(self.ch_param[ch])

            self.results.update({ch: dp_map})

            self.debug("dead pixel map: {}".format(dp_map))

            if output:
                fname = os.path.join(output, "dp_map_{}.csv".format(ch))
                ascii.write(
                    dp_map,
                    fname,
                    format="ecsv",
                    overwrite=True,
                    delimiter=",",
                )
                self.info("dead pixels map stored in {}".format(fname))

    def model(self, parameters):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`

        Returns
        --------
        :class:`~astropy.table.Table`
            channel dead pixel coordinates

        """

        # detector map
        y_size = parameters["detector"]["spatial_pix"]
        x_size = parameters["detector"]["spectral_pix"]

        if "dp_sigma" in parameters["detector"].keys():
            number_dead_pixels = RunConfig.random_generator.normal(
                loc=parameters["detector"]["dp_mean"],
                scale=parameters["detector"]["dp_sigma"],
                size=1,
            ).astype(int)
        else:
            number_dead_pixels = parameters["detector"]["dp_mean"]

        x_coords = RunConfig.random_generator.integers(
            0, x_size - 1, size=number_dead_pixels
        )
        y_coords = RunConfig.random_generator.integers(
            0, y_size - 1, size=number_dead_pixels
        )

        tab = Table()
        tab["spatial_coords"] = y_coords
        tab["spectral_coords"] = x_coords

        return tab



import astropy.units as u
import mpmath
import numpy as np

from exosim.tasks.task import Task
from exosim.utils.checks import check_units


class ComputeSolidAngle(Task):
    r"""
    It computes the solid angle given the system parameters.

    Returns
    --------
    :class:`~astropy.units.Quantity`
        solid angle expressed as :math:`sr \cdot m^2`
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        other_parameters: dict
            additional dict accounting for signal metadata
        """
        self.add_task_param("parameters", "channel parameters dict")
        self.add_task_param(
            "other_parameters",
            "additional dict accounting for signal metadata",
            None,
        )

    def execute(self):
        parameters = self.get_task_param("parameters")
        other_parameters = self.get_task_param("other_parameters")

        # all solid angles validated against ArielRad2 on 07 Sept 2021. Exact match.
        if other_parameters and "solid_angle" in other_parameters.keys():
            if other_parameters["solid_angle"] == "pi":
                omega_pix = np.pi * u.sr
            elif other_parameters["solid_angle"] == "pi-omega_pix":
                f_numx = parameters["Fnum_x"]
                if "Fnum_y" in parameters.keys():
                    f_numy = parameters["Fnum_y"]
                else:
                    f_numy = None
                omega_pix = np.pi * u.sr - self._omega_pix(f_numx, f_numy)
            elif isinstance(other_parameters["solid_angle"], u.Quantity):
                omega_pix = check_units(
                    other_parameters["solid_angle"], u.sr, self
                )
        else:
            f_numx = parameters["Fnum_x"]
            if "Fnum_y" in parameters.keys():
                f_numy = parameters["Fnum_y"]
            else:
                f_numy = None
            omega_pix = self._omega_pix(f_numx, f_numy)

        self.debug("omega pix: {}".format(omega_pix))

        area_pix = (parameters["detector"]["delta_pix"] ** 2).to(u.m**2)
        self.debug("pixel area: {}".format(area_pix))

        solid_angle = area_pix * omega_pix
        self.debug("solid angle: {}".format(solid_angle))

        self.set_output(solid_angle)

    def _omega_pix(self, Fnum_x, Fnum_y=None):
        """
        Calculate the solid angle subtended by an elliptical aperture on-axis.
        Algorithm from "John T. Conway. Nuclear Instruments and Methods in
        Physics Research Section A: Accelerators, Spectrometers, Detectors and
        Associated Equipment, 614(1):17 ? 27, 2010.
        https://doi.org/10.1016/j.nima.2009.11.075
        Equation n. 56

        Parameters
        ----------
        Fnum_x : scalar
                 Input F-number along dispersion direction
        Fnum_y : scalar
                 Optional, input F-number along cross-dispersion direction

        Returns
        -------
         Omega : scalar
                 The solid angle subtended by an elliptical aperture in units u.sr

        """

        if not Fnum_y:
            Fnum_y = Fnum_x

        if Fnum_x > Fnum_y:
            a = 1.0 / (2 * Fnum_y)
            b = 1.0 / (2 * Fnum_x)
        else:
            a = 1.0 / (2 * Fnum_x)
            b = 1.0 / (2 * Fnum_y)

        h = 1.0

        A = 4.0 * h * b / (a * np.sqrt(h * h + a * a))
        k = np.sqrt((a * a - b * b) / (h * h + a * a))
        alpha = np.sqrt(1 - (b / a) * (b / a))

        Omega = 2.0 * np.pi - A * mpmath.ellippi(alpha * alpha, k * k)

        return Omega * u.sr



import operator

import numpy as np

ops = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "%": operator.mod,
    "^": operator.xor,
}


def operate_over_axis(
    matrix: np.ndarray, vector: np.ndarray, axis: int, operation: str = "*"
) -> np.ndarray:
    """
    This function is used to perform an operation over a given axis of a matrix.

    Parameters
    ----------
    matrix : np.ndarray
        N dimensional array
    vector : np.ndarray
        1 dimensional array
    axis : int
        axis to operate over
    operation : str, optional
        operator symbol, by default "*"

    Returns
    -------
    np.ndarray
        operated matrix
    """
    axis_index = [np.newaxis] * matrix.ndim
    axis_index[axis] = slice(None)
    axis_index = tuple(axis_index)
    matrix = ops[operation](matrix, vector[axis_index])
    return matrix



from collections import OrderedDict

import numpy as np
from tqdm.auto import tqdm


def iterate_over_opticalElements(input, key, last_key, val):
    """
    It iterates over the optical element of a given class and returns the edited dictionary

    Parameters
    ----------
    input: dict
        input dictionaru
    key:
        optical element class
    last_key:
        optical element key to edit
    val:
        value to edit

    Returns
    -------
    dict
    """

    if key in input.keys():
        if "optical_path" in input[key].keys():
            _nested(input[key]["optical_path"], last_key, val)
        elif isinstance(input[key], OrderedDict):
            for ch in input[key].keys():
                if "optical_path" in input[key][ch].keys():
                    _nested(input[key][ch]["optical_path"], last_key, val)
                else:
                    _nested(input[key][ch], last_key, val)
        else:
            _nested(input[key], last_key, val)
    return input


def _nested(input, key, val):
    if isinstance(input["opticalElement"], OrderedDict):
        for opt in input["opticalElement"].keys():
            input["opticalElement"][opt][key] = val
    else:
        input["opticalElement"][key] = val


def iterate_over_chunks(dataset, **kwargs):
    """
    Iterates over the dataset chunks using tqdm to produce an adaptive progress bass
    Parameters
    ----------
    dataset: :class:`h5py.Dataset`
        h5py chunked dataset used to store the data

    Returns
    -------
        :class:`tqdm.tqdm`

    """
    total = int(np.ceil(dataset.shape[0] / dataset.chunks[0]))
    return tqdm(dataset.iter_chunks(), total=total, unit="chunk", **kwargs)


# from https://stackoverflow.com/a/20780569 on StackOverflow
# by cyborg (https://stackoverflow.com/users/907578/cyborg)
# this is compliant to StackOverflow's CC BY-SA 3.0
# and its attribution requirement
# (https://stackoverflow.blog/2009/06/25/attribution-required/)
def searchsorted(known_array, test_array):
    index_sorted = np.argsort(known_array)
    known_array_sorted = known_array[index_sorted]
    known_array_middles = (
        known_array_sorted[1:] - np.diff(known_array_sorted.astype("f")) / 2
    )
    idx1 = np.searchsorted(known_array_middles, test_array)
    indices = index_sorted[idx1]
    return indices



# This code is inspired by the code developed for ExoRad 2.
# Therefore, we attach here for ExoRad 2 license:
#
# BSD 3-Clause License
#
# Copyright (c) 2020, Lorenzo V. Mugnai, Enzo Pascale, "La Sapienza" Università di Roma
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the names of the copyright holders nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from typing import Tuple, Union

import astropy.units as u
import h5py
import matplotlib
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import numpy as np
from astropy.io.misc.hdf5 import read_table_hdf5
from astropy.table import Table
from photutils.aperture import EllipticalAperture, RectangularAperture

import exosim.log as log
from exosim.output.hdf5.utils import load_signal
from exosim.utils.ascii_arts import observatory

from .utils import _create_ordered_cmap, prepare_channels_list

plt.rcParams.update({"font.size": 16})


class RadiometricPlotter(log.Logger):
    """
    Radiometric plotter.
    This class handles the methods to plot the radiometric table produced by `exosim`.

    Attributes
    -----------
    input: str or :class:`astropy.table.QTable`
            input data
    input_table: :class:`astropy.table.QTable`
        input radiometric table
    fig: :class:`matplotlib.figure.Figure`
        produced figure

    Examples
    ----------
    The following example, given the `test_file.h5` preoduced by Exosim,
    plots the radiometric table and stores the figure as `radiometric.png`.

    >>> from exosim.plots import RadiometricPlotter
    >>> radiometricPlotter = RadiometricPlotter(input='./test_file.h5')
    >>> radiometricPlotter.plot_table()
    >>> radiometricPlotter.save_fig('radiometric.png')

    """

    def __init__(self, input: Union[str, Table]) -> None:
        """

        Parameters
        ----------
        input: str or :class:`astropy.table.QTable`
            input data
        """
        self.set_log_name()
        self.graphics(observatory)
        self.announce("started")

        self.input = input
        if isinstance(input, str):
            self.input_table = self.load_table(input)
        else:
            self.input_table = input
        self.fig = None

    def load_table(self, input_file: str) -> Table:
        """
        It loads the radiometric table from the input file:

        Parameters
        ----------
        input_file: str
            input file name

        Returns
        -------
        :class:`astropy.table.QTable`
            loaded radiometric table
        """
        with h5py.File(input_file, "r") as f:
            file_path = "radiometric"
            tab = read_table_hdf5(f[file_path], path="table")
        self.debug("radiometric table loaded")

        return tab

    def plot_bands(
        self,
        ax: plt.Axes,
        scale: str = "log",
        channel_edges: bool = True,
        add_legend: bool = True,
    ) -> plt.Axes:
        """
        It plots the channels bands behind the indicated axes.

        Parameters
        -----------
        ax: :class:`matplotlib.axes.Axes`
            axes where to plot the bands
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.

        Returns
        --------
        :class:`matplotlib.axes.Axes`
            axes with channel bands added

        """
        channels, norm = prepare_channels_list(self.input)
        cmap = _create_ordered_cmap("Pastel1", roll=-2, delete=-3)

        tick_list, patches = [], []
        for k, channel_name in enumerate(channels):
            wl_min = min(
                self.input_table["left_bin_edge"][
                    np.where(self.input_table["ch_name"] == channel_name)
                ]
            )
            if hasattr(wl_min, "unit"):
                wl_min = wl_min.value
            wl_max = max(
                self.input_table["right_bin_edge"][
                    np.where(self.input_table["ch_name"] == channel_name)
                ]
            )
            if hasattr(wl_max, "unit"):
                wl_max = wl_max.value
            ax.axvspan(
                wl_min,
                wl_max,
                alpha=0.3,
                zorder=0,
                color=cmap(
                    norm(k),
                ),
            )
            ax.axvspan(
                wl_min,
                wl_max,
                alpha=0.3,
                zorder=0,
                color=cmap(
                    norm(k),
                ),
            )
            #            wl_maxs += [wl_max]
            tick_list.append(wl_min)
            tick_list.append(wl_max)
            patches += [
                mpatches.Patch(
                    color=cmap(norm(k)), alpha=0.3, label=channel_name
                )
            ]
        #       tick_list.append(max(wl_maxs))
        ax.set_xscale(scale)
        if channel_edges:
            ax.set_xticks(tick_list)
            ax.get_xaxis().set_major_formatter(
                matplotlib.ticker.ScalarFormatter()
            )

        if add_legend:
            band_legend = ax.legend(
                handles=patches, title="Channels legend", bbox_to_anchor=(1, 1)
            )
            ax.add_artist(band_legend)
        return ax

    def plot_noise(
        self,
        ax: plt.Axes,
        scale: str = "log",
        channel_edges: bool = True,
        contribs: bool = False,
        ch_lengend: bool = True,
    ) -> plt.Axes:
        """
        It plots the noise components found in the input table in the indicated axes.

        Parameters
        -----------
        ax: :class:`matplotlib.axes.Axes`
            axes where to plot the noises
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.
        contribs: bool
            if ``True`` all the contributions are plotted. Default is ``False``.
        ch_lengend: bool
            if ``True`` add a legend for the channels color. Default is ``True``.

        Returns
        --------
        :class:`matplotlib.axes.Axes`
            axes with noises plotted

        """

        noise_keys = [
            x for x in self.input_table.keys() if "noise" in x or "custom" in x
        ]
        if not contribs:
            noise_keys = [k for k in noise_keys if "photon_noise" not in k]
            noise_keys += ["source_photon_noise", "foreground_photon_noise"]

        self.debug("noise keys : {}".format(noise_keys))
        for k, n in enumerate(noise_keys):
            if n == "total_noise":
                ax.plot(
                    self.input_table["Wavelength"],
                    self.input_table[n],
                    zorder=9,
                    lw=1,
                    c="k",
                    marker=".",
                    markersize=5,
                    label="total_noise",
                    alpha=0.8,
                )  # , c='None')
            else:
                if self.input_table[n].unit == u.hr**0.5:
                    noise = self.input_table[n]
                elif self.input_table[n].unit == u.ct / u.s:
                    self.debug(
                        "{} rescaled by starSignal_inAperture".format(n)
                    )
                    noise = (
                        self.input_table[n]
                        / self.input_table["source_signal_in_aperture"]
                        / (u.hr.to(u.s)) ** 0.5
                    )
                else:
                    self.error(
                        "{} unit not valid: {}".format(
                            n, self.input_table[n].unit
                        )
                    )
                # ax.scatter(self.input_table['Wavelength'], noise, label=n, zorder=10, s=5, color=palette[k])
                ax.plot(
                    self.input_table["Wavelength"],
                    noise,
                    zorder=9,
                    lw=1,
                    alpha=0.5,
                    marker=".",
                    label=n,
                )  # color=palette[k])  # c='None')

        #        ax.grid(zorder=0, which='both')
        locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        ax.yaxis.set_major_locator(locmaj)
        locmin = matplotlib.ticker.LogLocator(
            base=10.0, subs=(0.2, 0.4, 0.6, 0.8), numticks=12
        )
        ax.yaxis.set_minor_locator(locmin)
        ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.grid(axis="y", which="minor", alpha=0.3)
        ax.grid(axis="y", which="major", alpha=0.5)
        #        ax.legend(bbox_to_anchor=(1, 1))
        ax.set_title("Noise Budget")
        ax.set_xlabel(r"Wavelength [$\mu m$]")
        ax.set_ylabel(r"relative noise [$\sqrt{{hr}}$]")
        ax.set_yscale(scale)
        # ax.set_xscale('log')
        ax = self.plot_bands(ax, scale, channel_edges, add_legend=ch_lengend)
        ax.legend(
            prop={"size": 12},
            loc="upper left",
            ncol=3,
            bbox_to_anchor=(0.05, -0.25),
            labelspacing=1.2,
            handlelength=1,
        )
        return ax

    def plot_signal(
        self,
        ax: plt.Axes,
        ylim: Tuple[float, float] = None,
        scale: str = "log",
        channel_edges: bool = True,
        contribs: bool = False,
        ch_lengend: bool = True,
    ) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
        """
        It plots the signal components found in the input table in the indicated axes.

        Parameters
        -----------
        ylim: float or (float, float)
            ylim for :class:`matplotlib.axes.Axes`.
        ax: :class:`matplotlib.axes.Axes`
            axes where to plot the signals
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.
        contribs: bool
            if ``True`` all the contributions are plotted. Default is ``False``.
        ch_lengend: bool
            if ``True`` add a legend for the channels color. Default is ``True``.

        Returns
        --------
        :class:`matplotlib.axes.Axes`
            axes with signals plotted
        """

        keys = ["source_signal_in_aperture", "foreground_signal_in_aperture"]

        if contribs:
            keys = [
                x
                for x in self.input_table.keys()
                if "signal_in_aperture" in x and "noise" not in x
            ]

        self.debug("signal keys : {}".format(keys))
        for k, s in enumerate(keys):
            ax.plot(
                self.input_table["Wavelength"],
                self.input_table[s],
                zorder=9,
                lw=1,
                alpha=0.5,
                marker=".",
                label=s,
            )

        if ylim:
            ax.set_ylim(ylim)
        # elif ax.get_ylim()[0]<:
        #     ax.set_ylim(1e-3)
        #        ax.grid(zorder=0, which='both')
        locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        ax.yaxis.set_major_locator(locmaj)
        locmin = matplotlib.ticker.LogLocator(
            base=10.0, subs=(0.2, 0.4, 0.6, 0.8), numticks=12
        )
        ax.yaxis.set_minor_locator(locmin)
        ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        ax.grid(axis="y", which="minor", alpha=0.3)
        ax.grid(axis="y", which="major", alpha=0.5)
        #        ax.legend(bbox_to_anchor=(1, 1))
        ax.set_title("Signals")
        ax.set_xlabel(r"Wavelength [$\mu m$]")
        ax.set_ylabel("$ct/s$")
        ax.set_yscale(scale)

        ax = self.plot_bands(ax, scale, channel_edges, add_legend=ch_lengend)
        ax.legend(
            prop={"size": 12},
            loc="upper left",
            ncol=3,
            bbox_to_anchor=(0.05, -0.25),
            labelspacing=1.2,
            handlelength=1,
        )
        return ax

    def plot_table(
        self,
        scale: str = "log",
        channel_edges: bool = True,
        contribs: bool = False,
    ) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
        """
        It produces a figure with signal and noise for the input table.

        Parameters
        ----------
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.
        contribs: bool
            if ``True`` all the contributions are plotted. Default is ``False``.

        Returns
        --------
        :class:`matplotlib.figure.Figure`
            plotted figure
        (:class:`matplotlib.axes.Axes`, :class:`matplotlib.axes.Axes`)
            tuple of axis. First axes is for signal, second is for noise.

        """
        self.info("plotting radiometric table")
        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 10))
        # fig.suptitle(self.input_table.meta['name'])
        ax1 = self.plot_signal(
            ax1, scale=scale, channel_edges=channel_edges, contribs=contribs
        )
        ax2 = self.plot_noise(
            ax2,
            scale=scale,
            channel_edges=channel_edges,
            contribs=contribs,
            ch_lengend=False,
        )
        plt.tight_layout()
        plt.subplots_adjust(top=0.92, bottom=0.22, hspace=0.7, right=0.8)
        self.fig = fig
        return fig, (ax1, ax2)

    def plot_efficiency(
        self,
        scale: str = "log",
        channel_edges: bool = False,
        ch_lengend: bool = True,
    ) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
        """
        It produces a figure with efficiencies for the input table.

        Parameters
        ----------
        scale: str
            x axes scale. Default is `log`.
        channel_edges: bool
            if ``True`` the x axes ticks are placed at the channel edges. Default is ``True``.
        ch_lengend: bool
            if ``True`` add a legend for the channels color. Default is ``True``.
        Returns
        --------
        :class:`matplotlib.figure.Figure`
            plotted figure
        (:class:`matplotlib.axes.Axes`, :class:`matplotlib.axes.Axes`)
            tuple of axis. First axes is for signal, second is for noise.

        """
        self.info("plotting efficiency table")
        # fig.suptitle(self.input_table.meta['name'])
        channels = set(self.input_table["ch_name"])
        channels = list(channels)
        channels.sort()

        fig, (ax1, ax2) = plt.subplots(2, 1, figsize=(12, 10))
        with h5py.File(self.input, "r") as f:
            for ch in channels:
                eff_path = "channels/{}/efficiency".format(ch)
                eff = load_signal(f[eff_path])
                ax1.plot(eff.spectral, eff.data[0, 0], label=ch)

                resp_path = "channels/{}/responsivity".format(ch)
                resp = load_signal(f[resp_path])
                ax2.plot(resp.spectral, resp.data[0, 0], label=ch)

        # locmaj = matplotlib.ticker.LogLocator(base=10, numticks=12)
        # ax.yaxis.set_major_locator(locmaj)
        # locmin = matplotlib.ticker.LogLocator(base=10.0,
        #                                       subs=(0.2, 0.4, 0.6, 0.8),
        #                                       numticks=12)
        # ax.yaxis.set_minor_locator(locmin)
        # ax.yaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
        # ax.grid(axis='y', which='minor', alpha=0.3)
        # ax.grid(axis='y', which='major', alpha=0.5)
        # #        ax.legend(bbox_to_anchor=(1, 1))
        ax1.set_title("Efficiency")
        ax1.set_xlabel(r"Wavelength [$\mu m$]")
        # ax.set_ylabel('$ct/s$')
        # ax1.set_yscale(scale)
        ax2.set_title("Responsivity")
        ax2.set_ylabel(r"${}$".format(resp.data_units))
        ax2.set_xlabel(r"Wavelength [$\mu m$]")
        ax2.set_yscale("log")

        ax1 = self.plot_bands(ax1, scale, channel_edges, add_legend=ch_lengend)
        ax2 = self.plot_bands(ax2, scale, channel_edges, add_legend=ch_lengend)

        ax2.legend(
            prop={"size": 12},
            loc="upper left",
            ncol=7,
            bbox_to_anchor=(0.05, -0.25),
            labelspacing=1.2,
            handlelength=1,
        )
        plt.tight_layout()
        plt.subplots_adjust(
            top=0.92, bottom=0.12, hspace=0.7, right=0.8, left=0.1
        )
        self.fig = fig
        return fig, (ax1, ax2)

    def plot_apertures(self) -> plt.Figure:
        """
        It produces a figure with apertures superimposed to the focal plane.

        Returns
        --------
        :class:`matplotlib.figure.Figure`
            plotted figure

        """
        self.info("plotting apertures")
        table = self.load_table(self.input)

        def _prepare_figure():
            with h5py.File(self.input, "r") as f:
                ch_list = list(f["channels"].keys())
                ch_list.sort()
                widths = []
                for ch in ch_list:
                    size_x = f["channels"][ch]["focal_plane"][
                        "spectral"
                    ].shape[0]
                    size_y = f["channels"][ch]["focal_plane"]["spatial"].shape[
                        0
                    ]
                    widths += [int(np.ceil(size_x / size_y)), 0.1]

                heights = [1]
                scale = np.ceil(len(widths) / len(heights))
                size_y_fig = 10
                size_x_fig = size_y_fig * scale
                fig = plt.figure(
                    constrained_layout=True,
                    dpi=150,
                    figsize=(size_x_fig, size_y_fig),
                )

                spec = fig.add_gridspec(
                    ncols=len(widths),
                    nrows=len(heights),
                    width_ratios=widths,
                    height_ratios=heights,
                    wspace=0.1,
                    hspace=0.1,
                )
            return fig, spec, ch_list

        def _load_apertures(ch):
            center_spectral = table["spectral_center"][table["ch_name"] == ch]
            spectral_size = table["spectral_size"][table["ch_name"] == ch]
            center_spatial = table["spatial_center"][table["ch_name"] == ch]
            spatial_size = table["spatial_size"][table["ch_name"] == ch]
            shape = table["aperture_shape"][table["ch_name"] == ch]

            aperture_shapes = {
                "rectangular": RectangularAperture,
                "elliptical": EllipticalAperture,
            }

            aper = []
            for i in range(center_spectral.size):
                aper += [
                    aperture_shapes[shape[i]](
                        (center_spectral[i] - 1, center_spatial[i] - 1),
                        spectral_size[i],
                        spatial_size[i],
                    )
                ]

            return aper

        fig, spec, ch_list = _prepare_figure()
        i = 0
        for ch in ch_list:
            with h5py.File(self.input, "r") as f:
                file_path = os.path.join("channels", ch)
                focal_plane = load_signal(
                    f[os.path.join(file_path, "focal_plane")]
                )
                foreground = load_signal(
                    f[os.path.join(file_path, "frg_focal_plane")]
                )
                osf = focal_plane.metadata["oversampling"]

                final = focal_plane.data[0] + foreground.data[0]
                ax0 = fig.add_subplot(spec[0, i])

                im = ax0.imshow(
                    final[osf // 2 :: osf, osf // 2 :: osf],
                    interpolation="none",
                )
            ax0.set_title(ch)
            apertures = _load_apertures(ch)
            for aperture in apertures:
                aperture.plot(color="r", lw=2)

            i += 1

            plt.colorbar(im, ax=ax0, cax=fig.add_subplot(spec[i]))
            i += 1

        self.fig = fig
        return fig

    def save_fig(self, name: str) -> None:
        """
        It saves the produced figure.

        Parameters
        --------
        name: str
            figure name
        """
        dir_name = os.path.dirname(os.path.abspath(name))
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)

        try:
            self.fig.savefig("{}".format(name))

            self.info("plot saved in {}".format(name))
        except AttributeError:
            self.error(
                "the indicated figure is not available. Check if you have produced it."
            )



import logging
import os.path
import shutil

import exosim.tasks.load as load
from exosim.log import generate_logger_name
from exosim.utils.runConfig import RunConfig


def load_options(options_file):
    """
    It loads the configuration files into dictionaries.

    Parameters
    ----------
    options_file: str or dict
        configuration data to load

    Returns
    -------
    dict:
        main configuration dictionary
    dict:
        payload configuration dictionary

    """
    if isinstance(options_file, str):
        loadOption = load.LoadOptions()
        mainConfig = loadOption(filename=options_file)
    elif isinstance(options_file, dict):
        mainConfig = options_file
    payloadConfig = mainConfig["payload"]
    return mainConfig, payloadConfig


def copy_input_files(output_dir):
    """
    It copied the input configuration xml file to the output folder, if they are not there already.

    Parameters
    ----------
    output_dir: str
        output folder

    """
    for fname in RunConfig.config_file_list:
        try:
            shutil.copy(fname, output_dir)
            logger.debug(
                "{} copied in the destination folder".format(
                    os.path.basename(fname)
                )
            )
        except shutil.SameFileError:
            logger.debug(
                "{} already in the destination folder".format(
                    os.path.basename(fname)
                )
            )
            continue


def clean_config_files():
    """
    It clean the list of configuration files
    """
    RunConfig.config_file_list = []
    logger.debug(
        "Configuration files list cleaned: {}".format(
            RunConfig.config_file_list
        )
    )


logger = logging.getLogger(generate_logger_name(copy_input_files))



import h5py

import exosim.models.signal as signal
import exosim.output as output
from exosim.output.hdf5.utils import load_signal
from exosim.tasks.task import Task


class LoadQeMap(Task):
    """
    Loads the Quantum efficiency map

    Returns
    --------
    :class:`~exosim.models.signal.Signal`
        channel responsivity variation map

    Raises
    ------
    TypeError:
        if the output is not a :class:`~exosim.models.signal.Signal` class


    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary containing the parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid
        output: :class:`~exosim.output.output.Output` (optional)
            output file
        """

        self.add_task_param("parameters", "parameters dict")
        self.add_task_param("time", "time grid")
        self.add_task_param("output", "output file", None)

    def execute(self):
        self.info("loading quantum efficiency variation map")
        parameters = self.get_task_param("parameters")
        tt = self.get_task_param("time")

        qe_map = self.model(parameters, tt)

        # checking output
        if not isinstance(qe_map, signal.Signal):
            self.error("wrong output format")
            raise TypeError("wrong output format")

        self.debug("qe_variation_map: {}".format(qe_map.data))

        output_file = self.get_task_param("output")
        if output_file:
            if issubclass(output_file.__class__, output.Output):
                qe_map.write(output_file, "qe_variation_map")

        self.set_output(qe_map)

    def model(self, parameters, time):
        """

        Parameters
        ----------
        parameters: dict
            dictionary contained the channel parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Signal`
            channel responsivity

        """
        file_name = parameters["detector"]["qe_map_filename"]

        with h5py.File(file_name, "r") as f:
            qe = load_signal(f[parameters["value"]])

        return qe



import astropy.units as u
import numpy as np
from scipy.interpolate import interp1d

import exosim.utils as utils
from exosim.tasks.instrument.computeSourcesPointingOffset import angle_of_view
from exosim.tasks.task import Task
from exosim.utils.checks import check_units


class EstimateChJitter(Task):
    """
    It scales the pointing jitter expressed as :math:`deg` into the pixel unit according to the channel plate scale
    and interpolates it to a time grid which is aligned to the detector readout time grid.

    Returns
    --------
    (:class:`~numpy.ndarray`, :class:`~numpy.ndarray`)
        Tuple containing the pointing jitter in the spatial and spectral direction expressed in units of pixels.

    """

    def __init__(self):
        """
        Parameters
        __________
        pointing_jitter: (:class:`~astropy.units.Quantity`, :class:`~astropy.units.Quantity`,  :class:`~astropy.units.Quantity`)
            Tuple containing the pointing jitter in the spatial and spectral direction expressed in units of deg.
        parameters: dict
            dictionary containing the channel parameters.
            This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        """
        self.add_task_param("pointing_jitter", " ")
        self.add_task_param("parameters", "channel parameters dict")

    def execute(self):
        parameters = self.get_task_param("parameters")
        jitter_spa, jitter_spe, jitter_time = self.get_task_param(
            "pointing_jitter"
        )

        jitter_time = check_units(jitter_time, "s")
        jitter_time_step = jitter_time[1] - jitter_time[0]
        jitter_freq = check_units(jitter_time_step, "Hz")
        readout_freq = check_units(
            parameters["readout"]["readout_frequency"], "Hz"
        )

        new_freq = (
            np.ceil(jitter_freq.value / readout_freq.value) * readout_freq
        )

        new_step = check_units(new_freq, "s")

        new_jitter_time = utils.grids.time_grid(
            jitter_time[0], jitter_time[-1], new_step
        )
        new_jitter_time = new_jitter_time.to(u.s)

        jitter_spe_inter = interp1d(jitter_time, jitter_spe)
        jitter_spe = jitter_spe_inter(new_jitter_time) * jitter_spe.unit

        jitter_spa_inter = interp1d(jitter_time, jitter_spa)
        jitter_spa = jitter_spa_inter(new_jitter_time) * jitter_spa.unit

        aov_spatial, aov_spectral = angle_of_view(
            parameters["detector"]["plate_scale"],
            parameters["detector"]["delta_pix"],
            parameters["detector"]["oversampling"],
        )

        offset_spatial = jitter_spa.to(u.deg) / aov_spatial
        offset_spectral = jitter_spe.to(u.deg) / aov_spectral

        self.set_output(
            [
                jitter_spe,
                jitter_spa,
                offset_spatial,
                offset_spectral,
                new_jitter_time,
            ]
        )



import numpy as np
from astropy import units as u
from astropy.modeling.physical_models import BlackBody

from exosim.models import signal as signal
from exosim.tasks.task import Task
from exosim.utils import checks as checks


class LoadOpticalElement(Task):
    """
    Abstract class to load an optical element and returns the element self emission
    and it optical efficiency.


    Returns
    --------
    :class:`~exosim.models.signal.Radiance`
        optical element radiance
    :class:`~exosim.models.signal.Dimensionless`
        optical element efficiency

    Raises
    ------
    TypeError:
        if the outputs are not the right :class:`~exosim.models.signal.Signal` subclasses


    Notes
    -----
    This is a default class with standardised inputs and outputs.
    The user can load this class and overwrite the "model" method
    to implement a custom Task to replace this.
    """

    def __init__(self):
        """
        Parameters
        __________
        parameters: dict
            dictionary contained the optical element parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid
        """

        self.add_task_param("parameters", "optical element parameters dict")
        self.add_task_param("wavelength", "wavelength grid")
        self.add_task_param("time", "time grid")

    def execute(self):
        parameters = self.get_task_param("parameters")
        wl = self.get_task_param("wavelength")
        tt = self.get_task_param("time")

        radiance, efficiency = self.model(
            parameters=parameters, wavelength=wl, time=tt
        )

        # checking output
        if not isinstance(radiance, signal.Radiance):
            self.error("wrong output format")
            raise TypeError("wrong output format")
        if not isinstance(efficiency, signal.Dimensionless):
            self.error("wrong output format")
            raise u.UnitConversionError("wrong output format")

        self.set_output([radiance, efficiency])

    def model(self, parameters, wavelength, time):
        r"""
        It loads the indicated columns from the data file.

        If emissivity and temperature are provided the radiance is estimated as

        .. math::

            I_{surf}(\lambda) = \epsilon \cdot BB(\lambda, T)

        where :math:`\epsilon` is the indicated emissivity and :math:`BB(\lambda, T)` is the Planck black body law.


        Parameters
        ----------
        parameters: dict
            dictionary contained the sources parameters. This is usually parsed from :class:`~exosim.tasks.load.loadOptions.LoadOptions`
        wavelength: :class:`~astropy.units.Quantity`
            wavelength grid.
        time: :class:`~astropy.units.Quantity`
            time grid.

        Returns
        --------
        :class:`~exosim.models.signal.Radiance`
            optical element radiance
        :class:`~exosim.models.signal.Dimensionless`
            optical element efficiency

        """

        # load efficiency if provided
        if "efficiency_key" in parameters.keys():
            efficiency = self._get_data(
                parameters,
                wavelength,
                time,
                "efficiency_key",
                signal.Dimensionless,
            )
        # else we set it to 1
        else:
            efficiency = signal.Dimensionless(
                spectral=wavelength, data=np.ones(len(wavelength))
            )
            efficiency.temporal_rebin(time)

        # load radiance if provided
        if "radiance_key" in parameters.keys():
            radiance = self._get_data(
                parameters, wavelength, time, "radiance_key", signal.Radiance
            )
        # else we estimate it
        else:
            try:
                # if temperature is provided, we estimate the emission
                t_key = checks.find_key(
                    parameters.keys(), ["T", "temperature", "Temperature"]
                )
                T = parameters[t_key]
                T = checks.check_units(T, u.K)
                bb = BlackBody(T)
                bb_ = bb(wavelength).to(
                    u.W / u.m**2 / u.sr / u.um,
                    u.spectral_density(wavelength),
                )
            except KeyError:
                # if temperature is not provided, we consider zero emission
                bb_ = np.zeros(wavelength.size)

            radiance = signal.Radiance(spectral=wavelength, data=bb_)

            # load emissivity if provided
            if "emissivity_key" in parameters.keys():
                emissivity = self._get_data(
                    parameters,
                    wavelength,
                    time,
                    "emissivity_key",
                    signal.Dimensionless,
                )
                radiance *= emissivity

        self.debug("radiance: {}".format(radiance.data))

        return radiance, efficiency

    def _get_data(self, parameters, wl, tt, read_key, signal_type):
        # look for wavelength dependent data
        parsed_data = parameters["data"]
        self.debug("found emissivity file")

        wl_k = parameters["wavelength_key"]
        parsed_wl = checks.check_units(parsed_data[wl_k], "um")
        data_key = parameters[read_key]

        extracted_data = signal_type(
            spectral=parsed_wl, data=parsed_data[data_key].data
        )

        extracted_data.spectral_rebin(wl)
        extracted_data.temporal_rebin(tt)

        self.debug("emissivity: {}".format(extracted_data.data))
        return extracted_data



import logging

import numpy as np
from photutils.aperture import (
    EllipticalAperture,
    RectangularAperture,
    aperture_photometry,
)
from scipy.optimize import minimize

logger = logging.getLogger(__name__)


def find_rectangular_aperture(
    ima, desired_ene, center=None, start_h=None, start_w=None
):
    """
    Estimates the smallest rectangular aperture to collect the desired Encircled Energy from a PSF.
    It uses :func:`photutils.aperture.aperture_photometry`.

    Parameters
    ----------
    ima: :class:`~numpy.ndarray`
        two-dimensional array.
    desired_ene: float
        desired Encircled Energy
    center: (float, float)
        spectral and spatial coordinates of the aperture center in pixels.
        If None the center of the array is used. Default is `None`
    start_h: int
        starting aperture height
    start_w: int
        starting aperture width

    Returns
    -------
    (float, float)
        sizes of the rectangular aperture (h,w)
    float
        number of pixels in the aperture
    float
        encircled energy collected by the aperture

    Example
    --------
    In this example we find the optimal aperture containing the 85% of the energy of a PSF.

    >>> import matplotlib.pyplot as plt
    >>> import astropy.units as u
    >>> import photutils
    >>> from exosim.utils.aperture import find_rectangular_aperture
    >>> from exosim.utils.psf import create_psf
    >>>
    >>> img = create_psf(1*u.um, (60,40), 6*u.um)
    >>> size, area, ene = find_rectangular_aperture(img, 0.85)
    >>> positions = [(img.shape[1]//2,img.shape[0]//2)]
    >>> aperture = photutils.aperture.RectangularAperture(positions, size[0], size[1])
    >>>
    >>> plt.imshow(img)
    >>> aperture.plot(color='r', lw=2,)
    >>> plt.show()

    .. plot:: mpl_examples/find_rectangular_aperture.py

    """

    center = center if center else (ima.shape[1] / 2, ima.shape[0] / 2)

    def ene_func(center, h, w, ima):
        aper = RectangularAperture(center, h=h, w=w)
        phot_ = aperture_photometry(ima, aper)
        phot = phot_["aperture_sum"].data[0]
        return phot / ima.sum()

    def ene_fit(pos, center, ima, desired_ene):
        ene = ene_func(center, pos[0], pos[1], ima)
        enec = np.abs(ene - desired_ene)
        return enec

    start_h_ = start_h if start_h else 2.0
    start_w_ = start_w if start_w else 2.0

    res = minimize(
        ene_fit,
        x0=np.array([float(start_h_), float(start_w_)]),
        args=(center, ima, desired_ene),
        method="Nelder-Mead",
        tol=1.0e-6,
    )

    h = res.x[0]
    w = res.x[1]
    return (w, h), h * w, ene_func(center, h, w, ima)


def find_elliptical_aperture(ima, desired_ene, center=None):
    """
    Estimates the smallest elliptical aperture to collect the desired Encircled Energy from a PSF.
    It uses :func:`photutils.aperture.aperture_photometry`.

    Parameters
    ----------
    ima: :class:`~numpy.ndarray`
        two-dimensional array.
    desired_ene: float
        desired Encircled Energy
    center: (float, float)
        spectral and spatial coordinates of the aperture center in pixels.
        If None the center of the array is used. Default is `None`

    Returns
    -------
    (float, float)
        sizes of the elliptical aperture (h,w)
    float
        number of pixels in the aperture
    float
        encircled energy collected by the aperture

    Example
    --------

    In this example we find the optimal aperture containing the 85% of the energy of a PSF.

    >>> import matplotlib.pyplot as plt
    >>> import astropy.units as u
    >>> import photutils
    >>> from exosim.utils.psf import find_elliptical_aperture
    >>> from exosim.utils.psf import create_psf
    >>>
    >>> img = create_psf(1*u.um, (60,40), 6*u.um)
    >>> size, area, ene = find_elliptical_aperture(img, 0.85)
    >>> positions = [(img.shape[1]//2,img.shape[0]//2)]
    >>> aperture = photutils.aperture.EllipticalAperture(positions, size[0], size[1])
    >>>
    >>> plt.imshow(img)
    >>> aperture.plot(color='r', lw=2,)
    >>> plt.show()

    .. plot:: mpl_examples/find_elliptical_aperture.py


    """
    center = center if center else (ima.shape[1] / 2, ima.shape[0] / 2)

    def ene_func(center, a, b, ima):
        aper = EllipticalAperture(center, a=a, b=b)
        phot_ = aperture_photometry(ima, aper)
        phot = phot_["aperture_sum"].data[0]
        return phot / ima.sum()

    def ene_fit(pos, center, ima, desired_ene):
        ene = ene_func(center, pos[0], pos[1], ima)
        enec = np.abs(ene - desired_ene)
        return enec

    res = minimize(
        ene_fit,
        x0=np.array([1, 1]),
        args=(center, ima, desired_ene),
        method="Nelder-Mead",
        tol=1.0e-6,
    )

    a = res.x[0]
    b = res.x[1]
    return (a, b), a * b, ene_func(center, a, b, ima)


def find_bin_aperture(ima, desired_ene, spatial_with, center=None):
    """
    Estimates the smallest rectangular aperture for spectral bin of already fixed spatial size
    to collect the desired Encircled Energy from a PSF.
    It uses :func:`photutils.aperture.aperture_photometry`.

    Parameters
    ----------
    ima: :class:`~numpy.ndarray`
        two-dimensional array.
    spatial_with: float
        fixed bin spatial size
    desired_ene: float
        desired Encircled Energy
    center: (float, float)
        spectral and spatial coordinates of the aperture center in pixels.
        If None the center of the array is used. Default is `None`

    Returns
    -------
    float
        spectral size of the rectangular aperture
    float
        number of pixels in the aperture
    float
        encircled energy collected by the aperture
    """
    center = center if center else (ima.shape[1] / 2, ima.shape[0] / 2)

    def ene_func(center, h, w, psf):
        aper = RectangularAperture(center, h=h, w=w)
        phot_ = aperture_photometry(psf, aper)
        phot = phot_["aperture_sum"].data[0]

        aper_full = RectangularAperture(center, h=psf.shape[1], w=w)
        phot_ = aperture_photometry(psf, aper_full)
        phot_full = phot_["aperture_sum"].data[0]

        return phot / phot_full

    def ene_fit(pos, spatial_with, center, ima, desired_ene):
        h = pos[0]
        ene = ene_func(center, h, spatial_with, ima)
        enec = np.abs(ene - desired_ene)
        return enec

    res = minimize(
        ene_fit,
        x0=np.array([2.0]),
        args=(spatial_with, center, ima, desired_ene),
        method="Nelder-Mead",
        tol=1.0e-6,
    )

    h = res.x[0]
    return h, h * spatial_with, ene_func(center, h, spatial_with, ima)
